import tempfile
from typing import Any, Dict

from edam_ontology.streams import tabular_stream

from galaxy.tool_util.edam_util import load_edam_tree


def test_load_edam_tree():
    tree = load_edam_tree()
    _verify_tree(tree)


def test_load_edam_tree_from_path():
    tf = tempfile.NamedTemporaryFile("w")
    tf.write(tabular_stream().read())
    tf.flush()
    tree = load_edam_tree(tf.name)
    _verify_tree(tree)


def _verify_tree(tree: Dict[str, Any]):
    assert tree is not None
    assert "operation_0004" in tree
    assert "topic_3974" in tree
    assert tree["topic_3974"]["label"] == "Epistasis"
    assert tree["topic_3974"]["parents"] == ['topic_0622', 'topic_3295']
    assert tree["topic_3974"]["path"] == [['topic_3391', 'topic_0003'], ['topic_3070', 'topic_0003']]
    assert tree["topic_3974"]["definition"] == "The study of the epigenetic modifications of a whole cell, tissue, organism etc."



import collections
import gzip
import tempfile
from typing import Any, Dict, List, Optional, Tuple, Type

import pytest

from galaxy.tool_util.verify import (
    files_contains,
    files_delta,
    files_diff,
    files_re_match,
    files_re_match_multiline,
)


F1 = b"A\nB\nC"
F2 = b"A\nB\nD\nE" * 61
F3 = b"A\nB\n\xfc"
F4 = b"A\r\nB\nC"
MULTILINE_MATCH = b".*"
TestFile = collections.namedtuple('TestFile', 'value path')

TestDef = Tuple[bytes, bytes, Optional[Dict[str, Any]], Optional[Type[AssertionError]]]


def test_file_list():
    files = []
    for b, ext in [(F1, '.txt'), (F2, '.txt'), (F3, '.pdf'), (F4, '.txt'), (MULTILINE_MATCH, '.txt'), (F1, '.txt.gz')]:
        with tempfile.NamedTemporaryFile(mode='wb', suffix=ext, delete=False) as out:
            if ext == '.txt.gz':
                b = gzip.compress(b)
            out.write(b)
        files.append(TestFile(b, out.name))
    return files


def generate_tests(multiline=False):
    f1, f2, f3, f4, multiline_match, f5 = test_file_list()
    tests: List[TestDef]
    if multiline:
        tests = [(multiline_match, f1, {'lines_diff': 0, 'sort': True}, None)]
    else:
        tests = [(f1, f1, {'lines_diff': 0, 'sort': True}, None)]
    tests.extend([
        (f1, f2, {'lines_diff': 0, 'sort': True}, AssertionError),
        (f1, f3, None, AssertionError),
        (f1, f4, None, None),
        (f1, f5, {'decompress': True}, None),
    ])
    return tests


def generate_tests_sim_size():
    f1, f2, f3, f4, multiline_match, f5 = test_file_list()
    # tests for equal files
    tests: List[TestDef] = [
        (f1, f1, None, None),  # pass default values
        (f1, f1, {'delta': 0}, None),  # pass for values that should always pass
        (f1, f1, {'delta_frac': 0.0}, None),
    ]
    # tests for two different file (diff is 422 lines, 0.011709602)
    tests += [(f1, f2, None, None),  # pass default values
             (f1, f2, {'delta': 422}, None),  # test around the actual difference of two different
             (f1, f2, {'delta_frac': 84.4}, None),
             (f1, f2, {'delta': 421}, AssertionError),
             (f1, f2, {'delta_frac': 84.3}, AssertionError),
             (f1, f2, {'delta': 422, 'delta_frac': 84.3}, AssertionError),  # test with combination where at least one leads to an assertion
             (f1, f2, {'delta': 421, 'delta_frac': 84.4}, AssertionError)]
    return tests


@pytest.mark.parametrize('file1,file2,attributes,expect', generate_tests())
def test_files_contains(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_contains(file1.path, file2.path, attributes)
    else:
        files_contains(file2.path, file2.path, attributes)


@pytest.mark.parametrize('file1,file2,attributes,expect', generate_tests_sim_size())
def test_files_delta(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_delta(file1.path, file2.path, attributes)
    else:
        files_delta(file1.path, file2.path, attributes)


@pytest.mark.parametrize('file1,file2,attributes,expect', generate_tests())
def test_files_diff(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_diff(file1.path, file2.path, attributes)
    else:
        files_diff(file1.path, file2.path, attributes)


@pytest.mark.parametrize('file1,file2,attributes,expect', generate_tests())
def test_files_re_match(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_re_match(file1.path, file2.path, attributes)
    else:
        files_re_match(file1.path, file2.path, attributes)


@pytest.mark.parametrize('file1,file2,attributes,expect', generate_tests(multiline=True))
def test_files_re_match_multiline(file1, file2, attributes, expect):
    if expect is not None:
        with pytest.raises(expect):
            files_re_match_multiline(file1.path, file2.path, attributes)
    else:
        files_re_match_multiline(file1.path, file2.path, attributes)



from ._util import _assert_number


def assert_has_size(output_bytes, value: int = None, delta: int = 0, min: int = None, max: int = None, negate: bool = False):
    """
    Asserts the specified output has a size of the specified value,
    allowing for absolute (delta) and relative (delta_frac) difference.
    """
    output_size = len(output_bytes)
    _assert_number(output_size, value, delta, min, max, negate,
        "{expected} file size of {n}+-{delta}",
        "{expected} file size to be in [{min}:{max}]")



import os
import shutil
import tempfile

try:
    import h5py
except ImportError:
    h5py = None
import pytest

from galaxy.tool_util.parser.xml import __parse_assert_list_from_elem
from galaxy.tool_util.verify import asserts
from galaxy.util import etree

TABULAR_ASSERTION = """
    <assert_contents>
        <has_n_columns n="3"/>
    </assert_contents>
"""
TABULAR_CSV_ASSERTION = """
    <assert_contents>
        <has_n_columns sep="," min="3"/>
    </assert_contents>
"""
TABULAR_ASSERTION_COMMENT = """
    <assert_contents>
        <has_n_columns n="3" comment="#$"/>
    </assert_contents>
"""

TABULAR_DATA_POS = """1\t2\t3
"""

TABULAR_DATA_NEG = """1\t2\t3\t4
"""

TABULAR_CSV_DATA = """1,2
"""

TABULAR_DATA_COMMENT = """# comment
$ more comment (using a char with meaning wrt regexp)
1\t2\t3
"""

TEXT_HAS_TEXT_ASSERTION = """
    <assert_contents>
        <has_text text="test text"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N = """
    <assert_contents>
        <has_text text="test text" n="2"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_DELTA = """
    <assert_contents>
        <has_text text="test text" n="3" delta="1"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_MIN_MAX = """
    <assert_contents>
        <has_text text="test text" min="2" max="4"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_NEGATE = """
    <assert_contents>
        <has_text text="test text" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_NEGATE = """
    <assert_contents>
        <has_text text="test text" n="2" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE = """
    <assert_contents>
        <has_text text="test text" n="3" delta="1" negate="true"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE = """
    <assert_contents>
        <has_text text="test text" min="2" max="4" negate="true"/>
    </assert_contents>
"""

TEXT_NOT_HAS_TEXT_ASSERTION = """
    <assert_contents>
        <not_has_text text="not here"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_MATCHING_ASSERTION = """
    <assert_contents>
        <has_text_matching expression="te[sx]t"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_MATCHING_ASSERTION_N = """
    <assert_contents>
        <has_text_matching expression="te[sx]t" n="4"/>
    </assert_contents>
"""

TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX = """
    <assert_contents>
        <has_text_matching expression="te[sx]t" min="3" max="5"/>
    </assert_contents>
"""

TEXT_HAS_LINE_ASSERTION = """
    <assert_contents>
        <has_line line="test text"/>
    </assert_contents>
"""
TEXT_HAS_LINE_ASSERTION_N = """
    <assert_contents>
        <has_line line="test text" n="2"/>
    </assert_contents>
"""
TEXT_HAS_N_LINES_ASSERTION = """
    <assert_contents>
        <has_n_lines n="{n}"/>
    </assert_contents>
"""
TEXT_HAS_N_LINES_ASSERTION_DELTA = """
    <assert_contents>
        <has_n_lines n="{n}" delta="{delta}"/>
    </assert_contents>
"""
TEXT_HAS_LINE_MATCHING_ASSERTION = """
    <assert_contents>
        <has_line_matching expression="te[sx]t te[sx]t"/>
    </assert_contents>
"""
TEXT_HAS_LINE_MATCHING_ASSERTION_N = """
    <assert_contents>
        <has_line_matching expression="te[sx]t te[sx]t" n="2"/>
    </assert_contents>
"""

SIZE_HAS_SIZE_ASSERTION = """
    <assert_contents>
        <has_size value="{value}"/>
    </assert_contents>
"""
SIZE_HAS_SIZE_ASSERTION_DELTA = """
    <assert_contents>
        <has_size value="{value}" delta="{delta}"/>
    </assert_contents>
"""

TEXT_DATA_HAS_TEXT = """test text
"""

TEXT_DATA_HAS_TEXT_NEG = """desired content
is not here
"""

TEXT_DATA_NONE = None

TEXT_DATA_EMPTY = ""


XML_IS_VALID_XML_ASSERTION = """
    <assert_contents>
        <is_valid_xml/>
    </assert_contents>
"""
XML_HAS_ELEMENT_WITH_PATH = """
    <assert_contents>
        <has_element_with_path path="{path}"/>
    </assert_contents>
"""
XML_HAS_N_ELEMENTS_WITH_PATH = """
    <assert_contents>
        <is_valid_xml/>
        <has_n_elements_with_path path="{path}" n="{n}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT_MATCHES = """
    <assert_contents>
        <element_text_matches path="{path}" expression="{expression}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT_IS = """
    <assert_contents>
        <element_text_is path="{path}" text="{text}"/>
    </assert_contents>
"""
XML_ATTRIBUTE_MATCHES = """
    <assert_contents>
        <is_valid_xml/>
        <attribute_matches path="{path}" attribute="{attribute}" expression="{expression}"/>
    </assert_contents>
"""
XML_ELEMENT_TEXT = """
    <assert_contents>
        <element_text path="{path}">
            {content_assert}
        </element_text>
    </assert_contents>
"""

XML_XML_ELEMENT = """
    <assert_contents>
        <xml_element path="{path}" attribute="{attribute}" all="{all}" n="{n}" delta="{delta}" min="{min}" max="{max}" negate="{negate}">
            {content_assert}
        </xml_element>
    </assert_contents>
"""

VALID_XML = '''<root>
    <elem name="foo">
        <more name="bar">BAR</more>
        <more name="baz">BAZ</more>
        <more name="qux">QUX</more>
    </elem>
    <elem name="baz"/>
</root>
'''
INVALID_XML = '<root><elem name="foo"></root>'

if h5py is not None:
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        h5name = tmp.name
        with h5py.File(tmp.name, "w") as h5fh:
            h5fh.attrs['myfileattr'] = "myfileattrvalue"
            h5fh.attrs['myfileattrint'] = 1
            dset = h5fh.create_dataset("myint", (100,), dtype='i')
            dset.attrs['myintattr'] = "myintattrvalue"
            grp = h5fh.create_group("mygroup")
            grp.attrs['mygroupattr'] = "mygroupattrvalue"
            grp.create_dataset("myfloat", (50,), dtype='f')
            dset.attrs['myfloatattr'] = "myfloatattrvalue"
    with open(h5name, "rb") as h5fh:
        H5BYTES = h5fh.read()
    os.remove(h5name)

    H5_HAS_H5_KEYS = """
        <assert_contents>
            <has_h5_keys keys="myint,mygroup,mygroup/myfloat"/>
        </assert_contents>
    """
    H5_HAS_H5_KEYS_NEGATIVE = """
        <assert_contents>
            <has_h5_keys keys="absent"/>
        </assert_contents>
    """
    H5_HAS_ATTRIBUTE = """
        <assert_contents>
            <has_h5_attribute key="myfileattr" value="myfileattrvalue" />
            <has_h5_attribute key="myfileattrint" value="1" />
        </assert_contents>
    """
    H5_HAS_ATTRIBUTE_NEGATIVE = """
        <assert_contents>
            <has_h5_attribute key="myfileattr" value="wrong" />
            <has_h5_attribute key="myfileattrint" value="also_wrong" />
        </assert_contents>
    """

# create a test directory structure for zipping
# might also be done directly with the fipfile/tarfile module without creating
# a tmpdir, but its much harder to create empty directories or symlinks
tmpdir = tempfile.mkdtemp()
for f in ["file1.txt", "testdir/file1.txt", "testdir/file2.txt", "testdir/dir2/file1.txt"]:
    tmpfile = os.path.join(tmpdir, f)
    os.makedirs(os.path.dirname(tmpfile), exist_ok=True)
    with open(tmpfile, "w") as fh:
        fh.write(f)
os.makedirs(os.path.join(tmpdir, "emptydir"))
os.symlink("testdir/file1.txt", os.path.join(tmpdir, "symlink"))

with tempfile.NamedTemporaryFile(suffix=".zip", delete=False) as ziptmp:
    zipname = ziptmp.name
    shutil.make_archive(zipname[:-4], "zip", tmpdir)
    with open(zipname, "rb") as zfh:
        ZIPBYTES = zfh.read()
with tempfile.NamedTemporaryFile(suffix=".tar.gz", delete=False) as ziptmp:
    zipname = ziptmp.name
    shutil.make_archive(zipname[:-7], "gztar", tmpdir)
    with open(zipname, "rb") as zfh:
        TARBYTES = zfh.read()
shutil.rmtree(tmpdir)


with tempfile.NamedTemporaryFile(mode="w", delete=False) as nonarchivetmp:
    nonarchivename = nonarchivetmp.name
    nonarchivetmp.write("some text")
with open(nonarchivename, "rb") as ntmp:
    NONARCHIVE = ntmp.read()

ARCHIVE_HAS_ARCHIVE_MEMBER = """
    <assert_contents>
        <has_archive_member path="{path}" all="{all}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""

ARCHIVE_HAS_ARCHIVE_MEMBER_N = """
    <assert_contents>
        <has_archive_member path="{path}" n="{n}" delta="{delta}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""

ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX = """
    <assert_contents>
        <has_archive_member path="{path}" min="{min}" max="{max}">
            {content_assert}
        </has_archive_member>
    </assert_contents>
"""

TESTS = [
    # test successful assertion
    (
        TABULAR_ASSERTION, TABULAR_DATA_POS,
        lambda x: len(x) == 0
    ),
    # test wrong number of columns
    (
        TABULAR_ASSERTION, TABULAR_DATA_NEG,
        lambda x: 'Expected 3+-0 columns in output found 4' in x
    ),
    # test wrong number of columns for csv data
    (
        TABULAR_CSV_ASSERTION, TABULAR_CSV_DATA,
        lambda x: 'Expected the number of columns in output to be in [3:inf] found 2' in x
    ),
    # test tabular data with comments
    (
        TABULAR_ASSERTION_COMMENT, TABULAR_DATA_COMMENT,
        lambda x: len(x) == 0
    ),
    # test has_text
    (
        TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_text .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: "Expected text 'test text' in output ('desired content\nis not here\n')" in x
    ),
    # test has_text with None output
    (
        TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_NONE,
        lambda x: "Checking has_text assertion on empty output (None)" in x
    ),
    # test has_text with empty output
    (
        TEXT_HAS_TEXT_ASSERTION, TEXT_DATA_EMPTY,
        lambda x: "Expected text 'test text' in output ('')" in x
    ),
    # test has_text with n
    (
        TEXT_HAS_TEXT_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_text with n .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_N, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 2+-0 occurences of 'test text' in output ('test text\n') found 1" in x
    ),
    # test has_text with n and delta
    (
        TEXT_HAS_TEXT_ASSERTION_N_DELTA, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_text with n and delta .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_N_DELTA, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 3+-1 occurences of 'test text' in output ('test text\n') found 1" in x
    ),
    # test has_text with min max
    (
        TEXT_HAS_TEXT_ASSERTION_MIN_MAX, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_text with min max .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_MIN_MAX, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected that the number of occurences of 'test text' in output is in [2:4] ('test text\n') found 1" in x
    ),


    # test has_text negate
    (
        TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: len(x) == 0
    ),
    # test has_text negate .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_HAS_TEXT,
        lambda x: "Did not expect text 'test text' in output ('test text\n')" in x
    ),
    # test has_text negate with None output .. should have the same output as with negate="false"
    (
        TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_NONE,
        lambda x: "Checking has_text assertion on empty output (None)" in x
    ),
    # test has_text negate with empty output
    (
        TEXT_HAS_TEXT_ASSERTION_NEGATE, TEXT_DATA_EMPTY,
        lambda x: len(x) == 0
    ),
    # test has_text negate with n
    (
        TEXT_HAS_TEXT_ASSERTION_N_NEGATE, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_text negate with n .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_N_NEGATE, TEXT_DATA_HAS_TEXT * 2,
        lambda x: "Did not expect 2+-0 occurences of 'test text' in output ('test text\ntest text\n') found 2" in x
    ),
    # test has_text negate with n and delta
    (
        TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_text negate with n and delta .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_N_DELTA_NEGATE, TEXT_DATA_HAS_TEXT * 2,
        lambda x: "Did not expect 3+-1 occurences of 'test text' in output ('test text\ntest text\n') found 2" in x
    ),
    # test has_text negate with min max
    (
        TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_text negate with min max .. negative test
    (
        TEXT_HAS_TEXT_ASSERTION_MIN_MAX_NEGATE, TEXT_DATA_HAS_TEXT * 2,
        lambda x: "Did not expect that the number of occurences of 'test text' in output is in [2:4] ('test text\ntest text\n') found 2" in x
    ),


    # test not_has_text
    (
        TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test not_has_text .. negative test
    (
        TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: "Output file contains unexpected text 'not here'" in x
    ),
    # test not_has_text with None output
    (
        TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_NONE,
        lambda x: "Checking not_has_text assertion on empty output (None)" in x
    ),
    # test not_has_text with empty output
    (
        TEXT_NOT_HAS_TEXT_ASSERTION, TEXT_DATA_EMPTY,
        lambda x: len(x) == 0
    ),
    # test has_text_matching
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_text_matching .. negative test
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: "Expected text matching expression 'te[sx]t' in output ('desired content\nis not here\n')" in x
    ),
    # test has_text_matching with n
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_text_matching with n .. negative test (using the test text where "te[sx]st" appears twice)
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 4+-0 (non-overlapping) matches for 'te[sx]t' in output ('test text\n') found 2" in x
    ),
    # test has_text_matching with n
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_text_matching with n .. negative test (using the test text where "te[sx]st" appears twice)
    (
        TEXT_HAS_TEXT_MATCHING_ASSERTION_MINMAX, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected that the number of (non-overlapping) matches for 'te[sx]t' in output is in [3:5] ('test text\n') found 2" in x
    ),
    # test has_line
    (
        TEXT_HAS_LINE_ASSERTION, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_line .. negative test
    (
        TEXT_HAS_LINE_ASSERTION, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: "Expected line 'test text' in output ('desired content\nis not here\n')" in x
    ),
    # test has_line with n
    (
        TEXT_HAS_LINE_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_line with n .. negative test
    (
        TEXT_HAS_LINE_ASSERTION_N, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 2+-0 lines 'test text' in output ('test text\n') found 1" in x
    ),
    # test has_n_lines
    (
        TEXT_HAS_N_LINES_ASSERTION.format(n="2"), TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_n_lines .. bytes
    (
        TEXT_HAS_N_LINES_ASSERTION.format(n="2ki"), TEXT_DATA_HAS_TEXT * 2048,
        lambda x: len(x) == 0
    ),
    # test has_n_lines .. negative test
    (
        TEXT_HAS_N_LINES_ASSERTION.format(n="2"), TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 2+-0 lines in the output found 1" in x
    ),
    # test has_n_lines ..delta
    (
        TEXT_HAS_N_LINES_ASSERTION_DELTA.format(n="3", delta="1"), TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 3+-1 lines in the output found 1" in x
    ),
    # test has_line_matching
    (
        TEXT_HAS_LINE_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_line_matching .. negative test
    (
        TEXT_HAS_LINE_MATCHING_ASSERTION, TEXT_DATA_HAS_TEXT_NEG,
        lambda x: "Expected line matching expression 'te[sx]t te[sx]t' in output ('desired content\nis not here\n')" in x
    ),
    # test has_line_matching n
    (
        TEXT_HAS_LINE_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_line_matching n .. negative test
    (
        TEXT_HAS_LINE_MATCHING_ASSERTION_N, TEXT_DATA_HAS_TEXT,
        lambda x: "Expected 2+-0 lines matching for 'te[sx]t te[sx]t' in output ('test text\n') found 1" in x
    ),
    # test has_size
    (
        SIZE_HAS_SIZE_ASSERTION.format(value=10), TEXT_DATA_HAS_TEXT,
        lambda x: len(x) == 0
    ),
    # test has_size .. negative test
    (
        SIZE_HAS_SIZE_ASSERTION.format(value="10"), TEXT_DATA_HAS_TEXT * 2,
        lambda x: "Expected file size of 10+-0 found 20" in x
    ),
    # test has_size .. delta
    (
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(value="10", delta="10"), TEXT_DATA_HAS_TEXT * 2,
        lambda x: len(x) == 0
    ),
    # test has_size .. bytes suffix
    (
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(value="1k", delta="0"), TEXT_DATA_HAS_TEXT * 100,
        lambda x: len(x) == 0
    ),
    # test has_size .. bytes suffix .. negative
    (
        SIZE_HAS_SIZE_ASSERTION_DELTA.format(value="1Mi", delta="10k"), TEXT_DATA_HAS_TEXT * 100,
        lambda x: 'Expected file size of 1Mi+-10k found 1000' in x
    ),
    # test is_valid_xml
    (
        XML_IS_VALID_XML_ASSERTION, VALID_XML,
        lambda x: len(x) == 0
    ),
    # test is_valid_xml .. negative test
    (
        XML_IS_VALID_XML_ASSERTION, INVALID_XML,
        lambda x: 'Expected valid XML, but could not parse output. Opening and ending tag mismatch: elem line 1 and root, line 1, column 31 (<string>, line 1)' in x
    ),
    # test has_element_with_path
    (
        XML_HAS_ELEMENT_WITH_PATH.format(path="./elem[1]/more"), VALID_XML,
        lambda x: len(x) == 0
    ),
    (
        XML_HAS_ELEMENT_WITH_PATH.format(path="./elem[@name='foo']"), VALID_XML,
        lambda x: len(x) == 0
    ),
    (
        XML_HAS_ELEMENT_WITH_PATH.format(path=".//more[@name]"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test has_element_with_path .. negative test
    (
        XML_HAS_ELEMENT_WITH_PATH.format(path="./blah"), VALID_XML,
        lambda x: "Expected path './blah' in xml" in x
    ),
    # test has_n_elements_with_path
    (
        XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem", n="2"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test has_n_elements_with_path
    (
        XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[1]/more", n="3"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test has_n_elements_with_path
    (
        XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[@name='foo']/more", n="3"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test has_n_elements_with_path
    (
        XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem[2]/more", n="0"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test has_n_elements_with_path .. negative test
    (
        XML_HAS_N_ELEMENTS_WITH_PATH.format(path="./elem", n="1"), VALID_XML,
        lambda x: "Expected 1+-0 occurrences of path './elem' in xml found 2" in x
    ),
    # test element_text_matches
    (
        XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more", expression="BA(R|Z)"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text_matches more specific path
    (
        XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more[2]", expression="BA(R|Z)"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text_matches .. negative test
    (
        XML_ELEMENT_TEXT_MATCHES.format(path="./elem/more", expression="QU(X|Y)"), VALID_XML,
        lambda x: "Text of element with path './elem/more': Expected text matching expression 'QU(X|Y)' in output ('BAR')" in x
    ),
    # test element_text_is
    (
        XML_ELEMENT_TEXT_IS.format(path="./elem/more", text="BAR"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text_is with more specific path
    (
        XML_ELEMENT_TEXT_IS.format(path="./elem/more[@name='baz']", text="BAZ"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text_is .. negative test testing that prefix is not accepted
    (
        XML_ELEMENT_TEXT_IS.format(path="./elem/more", text="BA"), VALID_XML,
        lambda x: "Text of element with path './elem/more': Expected text matching expression 'BA$' in output ('BAR')" in x
    ),
    # test element_attribute_matches
    (
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more", attribute="name", expression="ba(r|z)"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_attribute_matches with more specific path
    (
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more[2]", attribute="name", expression="ba(r|z)"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_attribute_matches .. negative test
    (
        XML_ATTRIBUTE_MATCHES.format(path="./elem/more", attribute="name", expression="qu(x|y)"), VALID_XML,
        lambda x: "Attribute 'name' on element with path './elem/more': Expected text matching expression 'qu(x|y)' in output ('bar')" in x
    ),
    # test element_text
    (
        XML_ELEMENT_TEXT.format(path="./elem/more", content_assert=''), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text .. negative
    (
        XML_ELEMENT_TEXT.format(path="./absent", content_assert=''), VALID_XML,
        lambda x: "Expected path './absent' in xml" in x
    ),
    # test element_text with sub-assertion
    (
        XML_ELEMENT_TEXT.format(path="./elem/more", content_assert='<has_text text="BAR"/>'), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test element_text with sub-assertion .. negative
    (
        XML_ELEMENT_TEXT.format(path="./elem/more", content_assert='<has_text text="NOTBAR"/>'), VALID_XML,
        lambda x: "Text of element with path './elem/more': Expected text 'NOTBAR' in output ('BAR')" in x
    ),
    # note that xml_element is also tested indirectly by the other xml
    # assertions which are all implemented by xml_element
    # test xml_element
    (
        XML_XML_ELEMENT.format(path=".//more", n="2", delta="1", min="1", max="3", attribute="", all="false", content_assert='<has_text_matching expression="(BA[RZ]|QUX)$"/>', negate="false"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test xml_element testing attribute matching on all matching elements
    (
        XML_XML_ELEMENT.format(path=".//more", n="2", delta="1", min="1", max="3", attribute="name", all="true", content_assert='<has_text_matching expression="(ba[rz]|qux)$"/>', negate="false"), VALID_XML,
        lambda x: len(x) == 0
    ),
    # test xml_element .. failing because of n
    (
        XML_XML_ELEMENT.format(path=".//more", n="2", delta="0", min="1", max="3", attribute="", all="false", content_assert='', negate="false"), VALID_XML,
        lambda x: "Expected 2+-0 occurrences of path './/more' in xml found 3" in x
    ),
    # test xml_element .. failing because of n
    (
        XML_XML_ELEMENT.format(path=".//more", n="10000", delta="1", min="1", max="3", attribute="", all="false", content_assert='', negate="true"), VALID_XML,
        lambda x: "Did not expect that the number of occurences of path './/more' in xml is in [1:3] found 3" in x
    ),
    # test xml_element .. failing because of sub assertion
    (
        XML_XML_ELEMENT.format(path=".//more", n="2", delta="1", min="1", max="3", attribute="", all="false", content_assert='<has_text_matching expression="(BA[RZ]|QUX)$" negate="true"/>', negate="false"), VALID_XML,
        lambda x: "Text of element with path './/more': Did not expect text matching expression '(BA[RZ]|QUX)$' in output ('BAR')" in x
    ),

    # test has_archive_member with zip
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert="", all="false"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert="", all="false"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with non archive
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="irrelevant", content_assert="", all="false"), NONARCHIVE,
        lambda x: "Expected path 'irrelevant' to be an archive" in x
    ),
    # test has_archive_member with zip on absent member
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="absent", content_assert="", all="false"), ZIPBYTES,
        lambda x: "Expected path 'absent' in archive" in x
    ),
    # test has_archive_member with tar on absent member
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="absent", content_assert="", all="false"), TARBYTES,
        lambda x: "Expected path 'absent' in archive" in x
    ),
    # test has_archive_member with zip on symlink
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?symlink", content_assert='<has_text text="testdir/file1.txt"/>', all="false"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar on symlink
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?symlink", content_assert='<has_text text="testdir/file1.txt"/>', all="false"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip on a dir member (which are treated like empty files)
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/", content_assert='<has_size value="0"/>', all="false"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar on a dir member (which are treated like empty files)
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/", content_assert='<has_size value="0"/>', all="false"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip with subassertion (note that archive members are sorted therefor file1 in dir2 is tested)
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/.*\\.txt", content_assert='<has_text text="testdir/dir2/file1.txt"/>', all="false"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar with subassertion (note that archive members are sorted therefor file1 in dir2 is tested)
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/.*\\.txt", content_assert='<has_text text="testdir/dir2/file1.txt"/>', all="false"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip with failing subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert='<has_text text="ABSENT"/>', all="false"), ZIPBYTES,
        lambda x: "Archive member '(\\./)?testdir/file1.txt': Expected text 'ABSENT' in output ('testdir/file1.txt')" in x
    ),
    # test has_archive_member with tar with failing subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path="(\\./)?testdir/file1.txt", content_assert='<has_text text="ABSENT"/>', all="false"), TARBYTES,
        lambda x: "Archive member '(\\./)?testdir/file1.txt': Expected text 'ABSENT' in output ('testdir/file1.txt')" in x
    ),
    # test has_archive_member with zip checking all matches with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', all="true"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar checking all matches with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', all="true"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip checking all matches with failing subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file1\\.txt"/>', all="true"), ZIPBYTES,
        lambda x: "Expected text matching expression 'file1\\.txt' in output ('testdir/file2.txt')"
    ),
    # test has_archive_member with tar checking all matches with failing subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file1\\.txt"/>', all="true"), TARBYTES,
        lambda x: "Expected text matching expression 'file1\\.txt' in output ('testdir/file2.txt')"
    ),

    # test has_archive_member with zip n+delta with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="3", delta="1"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip n+delta with subassertion .. negative
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="1", delta="1"), ZIPBYTES,
        lambda x: "Expected 1+-1 matches for path '.*file.\\.txt' in archive found 4" in x
    ),
    # test has_archive_member with tar n+delta with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="3", delta="1"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar n+delta with subassertion .. negative
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_N.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', n="1", delta="1"), TARBYTES,
        lambda x: "Expected 1+-1 matches for path '.*file.\\.txt' in archive found 4" in x
    ),
    # test has_archive_member with zip min+max with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="2", max="4"), ZIPBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with zip min+max with subassertion .. negative
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="0", max="2"), ZIPBYTES,
        lambda x: "Expected that the number of matches for path '.*file.\\.txt' in archive is in [0:2] found 4" in x
    ),
    # test has_archive_member with tar min+max with subassertion
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="2", max="4"), TARBYTES,
        lambda x: len(x) == 0
    ),
    # test has_archive_member with tar min+max with subassertion .. negative
    (
        ARCHIVE_HAS_ARCHIVE_MEMBER_MINMAX.format(path=".*file.\\.txt", content_assert='<has_text_matching expression="file.\\.txt"/>', min="0", max="2"), TARBYTES,
        lambda x: "Expected that the number of matches for path '.*file.\\.txt' in archive is in [0:2] found 4" in x
    ),
]

if h5py is not None:
    H5PY_TESTS = [
        # test has_h5_keys
        (
            H5_HAS_H5_KEYS, H5BYTES,
            lambda x: len(x) == 0
        ),
        # test has_h5_keys .. negative
        (
            H5_HAS_H5_KEYS_NEGATIVE, H5BYTES,
            lambda x: "Not a HDF5 file or H5 keys missing:\n\t['mygroup', 'mygroup/myfloat', 'myint']\n\t['absent']" in x
        ),
        # test has_attribute
        (
            H5_HAS_ATTRIBUTE, H5BYTES,
            lambda x: len(x) == 0
        ),
        # test has_attribute .. negative
        (
            H5_HAS_ATTRIBUTE_NEGATIVE, H5BYTES,
            lambda x: "Not a HDF5 file or H5 attributes do not match:\n\t[('myfileattr', 'myfileattrvalue'), ('myfileattrint', 1)]\n\n\t(myfileattr : wrong)" in x
        ),
    ]
    TESTS.extend(H5PY_TESTS)


TEST_IDS = [
    'has_n_columns success',
    'has_n_columns failure',
    'has_n_columns for csv',
    'has_n_columns with comments',
    'has_text success',
    'has_text failure',
    'has_text None output',
    'has_text empty output',
    'has_text n success',
    'has_text n failure',
    'has_text n delta success',
    'has_text n delta failure',
    'has_text min/max delta success',
    'has_text min/max delta failure',
    'has_text negate success',
    'has_text negate failure',
    'has_text negate None output',
    'has_text negate empty output',
    'has_text negate n success',
    'has_text negate n failure',
    'has_text negate n delta success',
    'has_text negate n delta failure',
    'has_text negate min/max delta success',
    'has_text negate min/max delta failure',
    'not_has_text success',
    'not_has_text failure',
    'not_has_text None output',
    'not_has_text empty output',
    'has_text_matching success',
    'has_text_matching failure',
    'has_text_matching n success',
    'has_text_matching n failure',
    'has_text_matching min/max success',
    'has_text_matching min/max failure',
    'has_line success',
    'has_line failure',
    'has_line n success',
    'has_line n failure',
    'has_n_lines success',
    'has_n_lines n as bytes success',
    'has_n_lines failure',
    'has_n_lines delta',
    'has_line_matching success',
    'has_line_matching failure',
    'has_line_matching n success',
    'has_line_matching n failure',
    'has_size success',
    'has_size failure',
    'has_size delta',
    'has_size with bytes suffix',
    'has_size with bytes suffix failure',
    'is_valid_xml success',
    'is_valid_xml failure',
    'has_element_with_path success 1',
    'has_element_with_path success 2',
    'has_element_with_path success 3',
    'has_element_with_path failure',
    'has_n_elements_with_path success 1',
    'has_n_elements_with_path success 2',
    'has_n_elements_with_path success 3',
    'has_n_elements_with_path success 4',
    'has_n_elements_with_path failure',
    'element_text_matches sucess',
    'element_text_matches sucess (with more specific path)',
    'element_text_matches failure',
    'element_text_is sucess',
    'element_text_is sucess (with more specific path)',
    'element_text_is failure',
    'attribute_matches sucess',
    'attribute_matches sucess (with more specific path)',
    'attribute_matches failure',
    'element_text success',
    'element_text failure',
    'element_text with subassertion sucess',
    'element_text with subassertion failure',
    'xml_element matching text success',
    'xml_element matching attribute success',
    'xml_element failure (due to n)',
    'xml_element failure (due to min/max in combination with negate)',
    'xml_element failure (due to subassertion)',
    'has_archive_member zip',
    'has_archive_member tar',
    'has_archive_member non-archive',
    'has_archive_member zip absent member',
    'has_archive_member tar absent member',
    'has_archive_member zip symlink member',
    'has_archive_member tar symlink member',
    'has_archive_member zip non-file member',
    'has_archive_member tar non-file member',
    'has_archive_member zip with content assertion',
    'has_archive_member tar with content assertion',
    'has_archive_member zip with failing content assertion',
    'has_archive_member tar with failing content assertion',
    'has_archive_member zip all matching with content assertion',
    'has_archive_member tar all matching with content assertion',
    'has_archive_member zip all matching with failing content assertion',
    'has_archive_member tar all matching with failing content assertion',
    'has_archive_member zip n + delta and content assertion',
    'has_archive_member zip n + delta failing  and content assertion',
    'has_archive_member tar n + delta and content assertion',
    'has_archive_member tar n + delta failing and content assertion',
    'has_archive_member zip min max and content assertion',
    'has_archive_member zip min max failing  and content assertion',
    'has_archive_member tar min max and content assertion',
    'has_archive_member tar min max failing and content assertion',
]

if h5py is not None:
    H5PY_TEST_IDS = [
        'has_h5_keys',
        'has_h5_keys failure',
        'has_h5_attribute',
        'has_h5_attribute failure',
    ]
    TEST_IDS.extend(H5PY_TEST_IDS)


@pytest.mark.parametrize('assertion_xml,data,assert_func', TESTS, ids=TEST_IDS)
def test_assertions(assertion_xml, data, assert_func):
    assertion = etree.fromstring(assertion_xml)
    assertion_description = __parse_assert_list_from_elem(assertion)
    try:
        asserts.verify_assertions(data, assertion_description)
    except AssertionError as e:
        assert_list = e.args
    else:
        assert_list = ()
    assert assert_func(assert_list), assert_list



"""This module contains a linting functions for tool tests."""
from inspect import Parameter, signature

from ._util import is_datasource
from ..verify import asserts


# Misspelled so as not be picked up by nosetests.
def lint_tsts(tool_xml, lint_ctx):
    # determine node to report for general problems with tests
    tests = tool_xml.findall("./tests/test")
    general_node = tool_xml.find("./tests")
    if general_node is None:
        general_node = tool_xml.getroot()
    datasource = is_datasource(tool_xml)
    if not tests:
        if not datasource:
            lint_ctx.warn("No tests found, most tools should define test cases.", node=general_node)
        elif datasource:
            lint_ctx.info("No tests found, that should be OK for data_sources.", node=general_node)
        return

    num_valid_tests = 0
    for test_idx, test in enumerate(tests, start=1):
        has_test = False
        test_expect = ("expect_failure", "expect_exit_code", "expect_num_outputs")
        for te in test_expect:
            if te in test.attrib:
                has_test = True
                break
        test_assert = ("assert_stdout", "assert_stderr", "assert_command")
        for ta in test_assert:
            assertions = test.findall(ta)
            if len(assertions) == 0:
                continue
            if len(assertions) > 1:
                lint_ctx.error("Test {test_idx}: More than one {ta} found. Only the first is considered.")
            has_test = True
            _check_asserts(test_idx, assertions, lint_ctx)
        _check_asserts(test_idx, test.findall(".//assert_contents"), lint_ctx)

        # really simple test that test parameters are also present in the inputs
        for param in test.findall("param"):
            name = param.attrib.get("name", None)
            if not name:
                lint_ctx.error(f"Test {test_idx}: Found test param tag without a name defined.", node=param)
                continue
            name = name.split("|")[-1]
            xpaths = [f"@name='{name}'",
                      f"@argument='{name}'",
                      f"@argument='-{name}'",
                      f"@argument='--{name}'"]
            if "_" in name:
                xpaths += [f"@argument='-{name.replace('_', '-')}'",
                           f"@argument='--{name.replace('_', '-')}'"]
            found = False
            for xp in xpaths:
                inxpath = f".//inputs//param[{xp}]"
                inparam = tool_xml.findall(inxpath)
                if len(inparam) > 0:
                    found = True
                    break
            if not found:
                lint_ctx.error(f"Test {test_idx}: Test param {name} not found in the inputs", node=param)

        output_data_names, output_collection_names = _collect_output_names(tool_xml)
        found_output_test = False
        for output in test.findall("output") + test.findall("output_collection"):
            found_output_test = True
            name = output.attrib.get("name", None)
            if output.tag == "output":
                valid_names = output_data_names
            else:
                valid_names = output_collection_names
            if not name:
                lint_ctx.error(f"Test {test_idx}: Found {output.tag} tag without a name defined.", node=output)
            else:
                if name not in valid_names:
                    lint_ctx.error(f"Test {test_idx}: Found {output.tag} tag with unknown name [{name}], valid names [{valid_names}]", node=output)

        if "expect_failure" in test.attrib and found_output_test:
            lint_ctx.error(f"Test {test_idx}: Cannot specify outputs in a test expecting failure.", node=test)
            continue

        has_test = has_test or found_output_test
        if not has_test:
            lint_ctx.warn(f"Test {test_idx}: No outputs or expectations defined for tests, this test is likely invalid.", node=test)
        else:
            num_valid_tests += 1

    if num_valid_tests or datasource:
        lint_ctx.valid(f"{num_valid_tests} test(s) found.", node=general_node)
    else:
        lint_ctx.warn("No valid test(s) found.", node=general_node)


def _check_asserts(test_idx, assertions, lint_ctx):
    """
    assertions is a list of assert_contents, assert_stdout, assert_stderr, assert_command
    in practice only for the first case the list may be longer than one
    """
    for assertion in assertions:
        for i, a in enumerate(assertion.iter()):
            if i == 0:  # skip root note itself
                continue
            assert_function_name = "assert_" + a.tag
            if assert_function_name not in asserts.assertion_functions:
                lint_ctx.error(f"Test {test_idx}: unknown assertion '{a.tag}'", node=a)
                continue
            assert_function_sig = signature(asserts.assertion_functions[assert_function_name])
            # check type of the attributes (int, float ...)
            for attrib in a.attrib:
                if attrib not in assert_function_sig.parameters:
                    lint_ctx.error(f"Test {test_idx}: unknown attribute '{attrib}' for '{a.tag}'", node=a)
                    continue
                if assert_function_sig.parameters[attrib].annotation is not Parameter.empty:
                    try:
                        assert_function_sig.parameters[attrib].annotation(a.attrib[attrib])
                    except ValueError:
                        lint_ctx.error(
                            f"Test {test_idx}: attribute '{attrib}' for '{a.tag}' needs to be '{assert_function_sig.parameters[attrib].annotation.__name__}' got '{a.attrib[attrib]}'",
                            node=a)
            # check missing required attributes
            for p in assert_function_sig.parameters:
                if p in ["output", "output_bytes", "verify_assertions_function", "children"]:
                    continue
                if assert_function_sig.parameters[p].default is Parameter.empty and p not in a.attrib:
                    lint_ctx.error(f"Test {test_idx}: missing attribute '{p}' for '{a.tag}'", node=a)
            # has_n_lines, has_n_columns, and has_size need to specify n/value, min, or max
            if a.tag in ["has_n_lines", "has_n_columns"]:
                if "n" not in a.attrib and "min" not in a.attrib and "max" not in a.attrib:
                    lint_ctx.error(f"Test {test_idx}: '{a.tag}' needs to specify 'n', 'min', or 'max'", node=a)
            if a.tag == "has_size":
                if "value" not in a.attrib and "min" not in a.attrib and "max" not in a.attrib:
                    lint_ctx.error(f"Test {test_idx}: '{a.tag}' needs to specify 'n', 'min', or 'max'", node=a)


def _collect_output_names(tool_xml):
    output_data_names = []
    output_collection_names = []

    outputs = tool_xml.findall("./outputs")
    if len(outputs) == 1:
        for output in list(outputs[0]):
            name = output.attrib.get("name", None)
            if not name:
                continue
            if output.tag == "data":
                output_data_names.append(name)
            elif output.tag == "collection":
                output_collection_names.append(name)

    return output_data_names, output_collection_names



import os
import tempfile

import pytest

from galaxy.tool_util.lint import lint_tool_source_with, LintContext, XMLLintMessageLine, XMLLintMessageXPath
from galaxy.tool_util.linters import (
    citations,
    command,
    general,
    help,
    inputs,
    outputs,
    stdio,
    tests,
    xml_order,
)
from galaxy.tool_util.loader_directory import load_tool_sources_from_path
from galaxy.tool_util.parser.xml import XmlToolSource
from galaxy.util.xml_macros import load_with_references

# TODO tests tool xml for general linter
# tests tool xml for citations linter
CITATIONS_MULTIPLE = """
<tool>
    <citations/>
    <citations/>
</tool>
"""

CITATIONS_ABSENT = """
<tool/>
"""

CITATIONS_ERRORS = """
<tool>
    <citations>
        <nonsense/>
        <citation type="hoerensagen"/>
        <citation type="doi"> </citation>
    </citations>
</tool>
"""

CITATIONS_VALID = """
<tool>
    <citations>
        <citation type="doi">DOI</citation>
    </citations>
</tool>
"""

# tests tool xml for command linter
COMMAND_MULTIPLE = """
<tool>
    <command/>
    <command/>
</tool>
"""
COMMAND_MISSING = """
<tool/>
"""
COMMAND_TODO = """
<tool>
    <command>
        ## TODO
    </command>
</tool>
"""
COMMAND_DETECT_ERRORS_INTERPRETER = """
<tool>
    <command detect_errors="nonsense" interpreter="python"/>
</tool>
"""


# tests tool xml for general linter
GENERAL_MISSING_TOOL_ID_NAME_VERSION = """
<tool profile="2109">
</tool>
"""

GENERAL_WHITESPACE_IN_VERSIONS_AND_NAMES = """
<tool name=" BWA Mapper " id="bwa tool" version=" 1.0.1 " is_multi_byte="true" display_interface="true" require_login="true" hidden="true">
    <requirements>
        <requirement type="package" version=" 1.2.5 "> bwa </requirement>
    </requirements>
</tool>
"""

GENERAL_REQUIREMENT_WO_VERSION = """
<tool name="BWA Mapper" id="bwa_tool" version="1.0.1blah" is_multi_byte="true" display_interface="true" require_login="true" hidden="true" profile="20.09">
    <requirements>
        <requirement type="package">bwa</requirement>
        <requirement type="package" version="1.2.5"></requirement>
    </requirements>
</tool>
"""

GENERAL_VALID = """
<tool name="valid name" id="valid_id" version="1.0+galaxy1" profile="21.09">
</tool>
"""

# test tool xml for help linter
HELP_MULTIPLE = """
<tool>
    <help>Help</help>
    <help>More help</help>
</tool>
"""

HELP_ABSENT = """
<tool>
</tool>
"""

HELP_EMPTY = """
<tool>
    <help> </help>
</tool>
"""

HELP_TODO = """
<tool>
    <help>TODO</help>
</tool>
"""

HELP_INVALID_RST = """
<tool>
    <help>
        **xxl__
    </help>
</tool>
"""

# test tool xml for inputs linter
INPUTS_NO_INPUTS = """
<tool>
</tool>
"""

INPUTS_NO_INPUTS_DATASOURCE = """
<tool tool_type="data_source">
    <inputs/>
</tool>
"""

INPUTS_VALID = """
<tool>
    <inputs>
        <param name="txt_param" type="text"/>
        <param name="int_param" type="integer"/>
    </inputs>
</tool>
"""

INPUTS_PARAM_NAME = """
<tool>
    <inputs>
        <param type="text"/>
        <param name="" type="text"/>
        <param name="2" type="text"/>
        <param argument="--valid" type="text"/>
        <param name="param_name" argument="--param-name" type="text"/>
    </inputs>
</tool>
"""

INPUTS_PARAM_TYPE = """
<tool>
    <inputs>
        <param name="valid_name"/>
        <param argument="--another-valid-name" type=""/>
    </inputs>
</tool>
"""

INPUTS_DATA_PARAM = """
<tool>
    <inputs>
        <param name="valid_name" type="data"/>
    </inputs>
</tool>
"""

INPUTS_CONDITIONAL = """
<tool>
    <inputs>
        <conditional>
            <param name="select" type="select"/>
        </conditional>
        <conditional name="cond_wo_param">
        </conditional>
        <conditional name="cond_w_mult_param">
            <param name="select3" type="select"><option value="A">A</option><option value="B">B</option></param>
            <param name="select4" type="select"><option value="A">A</option><option value="B">B</option></param>
            <when value="A"/>
            <when value="B"/>
        </conditional>
        <conditional name="cond_boolean">
            <param name="bool" type="boolean"/>
            <when value="true"/>
            <when value="false"/>
            <when value="False"/>
        </conditional>
        <conditional name="cond_text">
            <param name="text" type="text"/>
        </conditional>
        <conditional name="cond_w_optional_select">
            <param name="optionalselect" type="select" optional="true"><option value="A">A</option><option value="B">B</option></param>
            <when value="A"/>
            <when value="B"/>
        </conditional>
        <conditional name="cond_w_multiple_select">
            <param name="multipleselect" type="select" multiple="true"><option value="A">A</option><option value="B">B</option></param>
            <when value="A"/>
            <when value="B"/>
        </conditional>
        <conditional name="when_wo_value">
            <param name="select3" type="select"><option value="A">A</option><option value="B">B</option></param>
            <when/>
            <when value="A"/>
            <when value="B"/>
        </conditional>
        <conditional name="missing_when">
            <param name="label_select" type="select">
                <option value="none" selected="True">None</option>
            </param>
        </conditional>
        <conditional name="missing_option">
            <param name="missing_option" type="select">
                <option value="none" selected="True">None</option>
            </param>
            <when value="none"/>
            <when value="absent"/>
        </conditional>
    </inputs>
</tool>
"""

INPUTS_SELECT_INCOMPATIBLE_DISPLAY = """
<tool>
    <inputs>
        <param name="radio_select" type="select" display="radio" optional="true" multiple="true">
            <option value="1">1</option>
            <option value="2">2</option>
        </param>
        <param name="checkboxes_select" type="select" display="checkboxes" optional="false" multiple="false">
            <option value="1">1</option>
            <option value="2">2</option>
        </param>
        <!-- this must not raise any warning/error since multiple=true implies true as default for optional -->
        <param name="checkboxes_select_correct" type="select" display="checkboxes" multiple="true">
            <option value="1">1</option>
            <option value="2">2</option>
        </param>
    </inputs>
</tool>
"""

INPUTS_SELECT_DUPLICATED_OPTIONS = """
<tool>
    <inputs>
        <param name="select" type="select" optional="true" multiple="true">
            <option value="v">x</option>
            <option value="v">x</option>
        </param>
    </inputs>
</tool>
"""

SELECT_DUPLICATED_OPTIONS_WITH_DIFF_SELECTED = """
<tool>
    <inputs>
        <param name="select" type="select" optional="true" multiple="true">
            <option value="v">x</option>
            <option value="v" selected="true">x</option>
        </param>
    </inputs>
</tool>
"""

INPUTS_SELECT_DEPRECATIONS = """
<tool>
    <inputs>
        <param name="select_do" type="select" dynamic_options="blah()"/>
        <param name="select_ff" type="select">
            <options from_file="file.tsv" transform_lines="narf()"/>
        </param>
        <param name="select_fp" type="select">
            <options from_parameter="select_do" options_filter_attribute="fasel"/>
        </param>
    </inputs>
</tool>
"""

INPUTS_SELECT_OPTION_DEFINITIONS = """
<tool>
    <inputs>
        <param name="select_noopt" type="select"/>
        <param name="select_noopts" type="select">
            <options/>
        </param>
        <param name="select_fd_op" type="select">
            <options from_dataset="xyz"/>
            <options from_data_table="xyz"/>
            <option value="x">x</option>
        </param>
        <param name="select_fd_fdt" type="select">
            <options from_dataset="xyz" from_data_table="xyz"/>
        </param>
        <param name="select_noval_notext" type="select">
            <option>option wo value</option>
            <option value="value"/>
        </param>
        <param name="select_meta_file_key_incomp" type="select">
            <options from_data_table="xyz" meta_file_key="dbkey"/>
        </param>
    </inputs>
</tool>
"""

INPUTS_SELECT_FILTER = """
<tool>
    <inputs>
        <param name="select_filter_types" type="select">
            <options from_data_table="xyz">
                <filter/>
                <filter type="unknown_filter_type"/>
            </options>
        </param>
    </inputs>
</tool>
"""

INPUTS_VALIDATOR_INCOMPATIBILITIES = """
<tool>
    <inputs>
        <param name="param_name" type="text">
            <validator type="in_range">TEXT</validator>
            <validator type="regex" filename="blah"/>
            <validator type="expression"/>
            <validator type="expression">[</validator>
            <validator type="value_in_data_table"/>
        </param>
        <param name="another_param_name" type="data" format="bed">
            <validator type="metadata"/>
        </param>
    </inputs>
</tool>
"""

INPUTS_VALIDATOR_CORRECT = """
<tool>
    <inputs>
        <param name="data_param" type="data" format="data">
            <validator type="metadata" check="md1,md2" skip="md3,md4" message="cutom validation message" negate="true"/>
            <validator type="unspecified_build" message="cutom validation message" negate="true"/>
            <validator type="dataset_ok_validator" message="cutom validation message" negate="true"/>
            <validator type="dataset_metadata_in_range" min="0" max="100" exclude_min="true" exclude_max="true" message="cutom validation message" negate="true"/>
            <validator type="dataset_metadata_in_file" filename="file.tsv" metadata_column="3" split=","  message="cutom validation message" negate="true"/>
            <validator type="dataset_metadata_in_data_table" table_name="datatable_name" metadata_column="3" message="cutom validation message" negate="true"/>
        </param>
        <param name="collection_param" type="collection">
            <validator type="metadata" check="md1,md2" skip="md3,md4" message="cutom validation message"/>
            <validator type="unspecified_build" message="cutom validation message"/>
            <validator type="dataset_ok_validator" message="cutom validation message"/>
            <validator type="dataset_metadata_in_range" min="0" max="100" exclude_min="true" exclude_max="true" message="cutom validation message"/>
            <validator type="dataset_metadata_in_file" filename="file.tsv" metadata_column="3" split=","  message="cutom validation message"/>
            <validator type="dataset_metadata_in_data_table" table_name="datatable_name" metadata_column="3" message="cutom validation message"/>
        </param>
        <param name="text_param" type="text">
            <validator type="regex">reg.xp</validator>
            <validator type="length" min="0" max="100" message="cutom validation message"/>
            <validator type="empty_field" message="cutom validation message"/>
            <validator type="value_in_data_table" table_name="datatable_name" metadata_column="3" message="cutom validation message"/>
            <validator type="expression" message="cutom validation message">somepythonexpression</validator>
        </param>
        <param name="select_param" type="select">
            <options from_data_table="bowtie2_indexes"/>
            <validator type="no_options" negate="true"/>
            <validator type="regex" negate="true">reg.xp</validator>
            <validator type="length" min="0" max="100" message="cutom validation message" negate="true"/>
            <validator type="empty_field" message="cutom validation message" negate="true"/>
            <validator type="value_in_data_table" table_name="datatable_name" metadata_column="3" message="cutom validation message" negate="true"/>
            <validator type="expression" message="cutom validation message" negate="true">somepythonexpression</validator>
        </param>
        <param name="int_param" type="integer">
            <validator type="in_range" min="0" max="100" exclude_min="true" exclude_max="true" negate="true"/>
            <validator type="expression" message="cutom validation message">somepythonexpression</validator>
        </param>
    </inputs>
</tool>
"""


# test tool xml for outputs linter
OUTPUTS_MISSING = """
<tool/>
"""
OUTPUTS_MULTIPLE = """
<tool>
    <outputs/>
    <outputs/>
</tool>
"""
OUTPUTS_UNKNOWN_TAG = """
<tool>
    <outputs>
        <output/>
    </outputs>
</tool>
"""
OUTPUTS_UNNAMED_INVALID_NAME = """
<tool>
    <outputs>
        <data/>
        <collection name="2output"/>
    </outputs>
</tool>
"""
OUTPUTS_FORMAT_INPUT = """
<tool>
    <outputs>
        <data name="valid_name" format="input"/>
    </outputs>
</tool>
"""

# check that linter accepts format source for collection elements as means to specify format
# and that the linter warns if format and format_source are used
OUTPUTS_COLLECTION_FORMAT_SOURCE = """
<tool>
    <outputs>
        <collection name="output_collection" type="paired">
            <data name="forward" format_source="input_readpair" />
            <data name="reverse" format_source="input_readpair" format="fastq"/>
        </collection>
    </outputs>
</tool>
"""

# check that setting format with actions is supported
OUTPUTS_FORMAT_ACTION = """
<tool>
    <outputs>
        <data name="output">
            <actions>
                <conditional name="library.type">
                    <when value="paired">
                        <action type="format">
                            <option type="from_param" name="library.input_2" param_attribute="ext" />
                        </action>
                    </when>
                </conditional>
            </actions>
        </data>
    </outputs>
</tool>
"""

# check that linter does not complain about missing format if from_tool_provided_metadata is used
OUTPUTS_DISCOVER_TOOL_PROVIDED_METADATA = """
<tool>
    <outputs>
        <data name="output">
            <discover_datasets from_tool_provided_metadata="true"/>
        </data>
    </outputs>
</tool>
"""

# tool xml for repeats linter
REPEATS = """
<tool>
    <inputs>
        <repeat>
            <param name="another_param_name" type="data" format="bed"/>
        </repeat>
    </inputs>
</tool>
"""

# tool xml for stdio linter
STDIO_DEFAULT_FOR_DEFAULT_PROFILE = """
<tool>
</tool>
"""

STDIO_DEFAULT_FOR_NONLEGACY_PROFILE = """
<tool profile="21.09">
</tool>
"""

STDIO_MULTIPLE_STDIO = """
<tool>
    <stdio/>
    <stdio/>
</tool>
"""

STDIO_INVALID_CHILD_OR_ATTRIB = """
<tool>
    <stdio>
        <reqex/>
        <regex descriptio="blah" level="fatal" match="error" source="stdio"/>
        <exit_code descriptio="blah" level="fatal" range="1:"/>
    </stdio>
</tool>
"""

STDIO_INVALID_MATCH = """
<tool>
    <stdio>
        <regex match="["/>
    </stdio>
</tool>
"""

# check that linter does complain about tests wo assumptions
TESTS_ABSENT = """
<tool/>
"""
TESTS_ABSENT_DATA_SOURCE = """
<tool tool_type="data_source"/>
"""
TESTS_WO_EXPECTATIONS = """
<tool>
    <tests>
        <test>
        </test>
    </tests>
</tool>
"""

TESTS_PARAM_OUTPUT_NAMES = """
<tool>
    <inputs>
        <param argument="--existent-test-name"/>
        <conditional>
            <when>
                <param name="another_existent_test_name"/>
            </when>
        </conditional>
    </inputs>
    <outputs>
        <data name="existent_output"/>
        <collection name="existent_collection"/>
    </outputs>
    <tests>
        <test expect_num_outputs="1">
            <param/>
            <param name="existent_test_name"/>
            <param name="cond_name|another_existent_test_name"/>
            <param name="non_existent_test_name"/>
            <output/>
            <output name="existent_output"/>
            <output name="nonexistent_output"/>
            <output_collection/>
            <output_collection name="existent_collection"/>
            <output_collection name="nonexistent_collection"/>
        </test>
    </tests>
</tool>
"""

TESTS_EXPECT_FAILURE_OUTPUT = """
<tool>
    <outputs>
        <data name="test"/>
    </outputs>
    <tests>
        <test expect_failure="true">
            <output name="test"/>
        </test>
    </tests>
</tool>
"""

ASSERTS = """
<tool>
    <outputs>
        <data name="out_archive"/>
        <data name="out_tabular"/>
    </outputs>
    <tests>
        <test>
            <assert_stdout>
                <has_text text="blah" n="1"/>
                <has_line line="blah"/>
            </assert_stdout>
            <assert_stderr>
                <invalid/>
            </assert_stderr>
            <assert_command>
                <has_text invalid_attrib="blah"/>
            </assert_command>
            <output name="out_archive">
                <assert_contents>
                    <has_size value="500k" delta="1O"/>
                    <has_archive_member path=".*/my-file.txt">
                        <not_has_text invalid_attrib_also_checked_in_nested_asserts="Blah" text="EDK72998.1" />
                    </has_archive_member>
                </assert_contents>
            </output>
            <output name="out_tabular">
                <assert_contents>
                    <has_size/>
                    <has_n_columns/>
                    <has_n_lines/>
                </assert_contents>
            </output>
        </test>
    </tests>
</tool>
"""
TESTS_VALID = """
<tool>
    <outputs>
        <data name="test"/>
    </outputs>
    <tests>
        <test>
            <output name="test"/>
        </test>
    </tests>
</tool>
"""

# tool xml for xml_order linter
XML_ORDER = """
<tool>
    <wrong_tag/>
    <command/>
    <stdio/>
</tool>
"""


@pytest.fixture()
def lint_ctx():
    return LintContext('all', lint_message_class=XMLLintMessageLine)


@pytest.fixture()
def lint_ctx_xpath():
    return LintContext('all', lint_message_class=XMLLintMessageXPath)


def get_xml_tool_source(xml_string):
    with tempfile.NamedTemporaryFile(mode="w", suffix="tool.xml") as tmp:
        tmp.write(xml_string)
        tmp.flush()
        tool_path = tmp.name
        return load_with_references(tool_path)[0]


def failed_assert_print(lint_ctx):
    return (
        f"Valid: {lint_ctx.valid_messages}\n"
        f"Info: {lint_ctx.info_messages}\n"
        f"Warnings: {lint_ctx.warn_messages}\n"
        f"Errors: {lint_ctx.error_messages}"
    )


def run_lint(lint_ctx, lint_func, lint_target):
    lint_ctx.lint(name="test_lint", lint_func=lint_func, lint_target=lint_target)
    # check if the lint messages have the line
    for message in lint_ctx.message_list:
        if lint_func != general.lint_general:
            assert message.line is not None, f"No context found for message: {message.message}"


def test_citations_multiple(lint_ctx):
    tool_source = get_xml_tool_source(CITATIONS_MULTIPLE)
    run_lint(lint_ctx, citations.lint_citations, tool_source)
    assert "More than one citation section found, behavior undefined." in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 1


def test_citations_absent(lint_ctx):
    tool_source = get_xml_tool_source(CITATIONS_ABSENT)
    run_lint(lint_ctx, citations.lint_citations, tool_source)
    assert lint_ctx.warn_messages == ["No citations found, consider adding citations to your tool."]
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.error_messages


def test_citations_errors(lint_ctx):
    tool_source = get_xml_tool_source(CITATIONS_ERRORS)
    run_lint(lint_ctx, citations.lint_citations, tool_source)
    assert "Unknown tag discovered in citations block [nonsense], will be ignored." in lint_ctx.warn_messages
    assert "Unknown citation type discovered [hoerensagen], will be ignored." in lint_ctx.warn_messages
    assert 'Empty doi citation.' in lint_ctx.error_messages
    assert 'Found no valid citations.' in lint_ctx.warn_messages
    assert len(lint_ctx.warn_messages) == 3
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages


def test_citations_valid(lint_ctx):
    tool_source = get_xml_tool_source(CITATIONS_VALID)
    run_lint(lint_ctx, citations.lint_citations, tool_source)
    assert 'Found 1 likely valid citations.' in lint_ctx.valid_messages
    assert len(lint_ctx.valid_messages) == 1
    assert not lint_ctx.info_messages
    assert not lint_ctx.error_messages


def test_command_multiple(lint_ctx):
    tool_source = get_xml_tool_source(COMMAND_MULTIPLE)
    run_lint(lint_ctx, command.lint_command, tool_source)
    assert 'More than one command tag found, behavior undefined.' in lint_ctx.error_messages
    assert len(lint_ctx.error_messages) == 1
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages


def test_command_missing(lint_ctx):
    tool_source = get_xml_tool_source(COMMAND_MISSING)
    run_lint(lint_ctx, command.lint_command, tool_source)
    assert 'No command tag found, must specify a command template to execute.' in lint_ctx.error_messages


def test_command_todo(lint_ctx):
    tool_source = get_xml_tool_source(COMMAND_TODO)
    run_lint(lint_ctx, command.lint_command, tool_source)
    assert 'Tool contains a command.' in lint_ctx.info_messages
    assert 'Command template contains TODO text.' in lint_ctx.warn_messages


def test_command_detect_errors_interpreter(lint_ctx):
    tool_source = get_xml_tool_source(COMMAND_DETECT_ERRORS_INTERPRETER)
    run_lint(lint_ctx, command.lint_command, tool_source)
    assert "Command uses deprecated 'interpreter' attribute." in lint_ctx.warn_messages
    assert 'Tool contains a command with interpreter of type [python].' in lint_ctx.info_messages
    assert 'Unknown detect_errors attribute [nonsense]' in lint_ctx.warn_messages
    assert 'Command is empty.' in lint_ctx.error_messages


def test_general_missing_tool_id_name_version(lint_ctx):
    tool_source = get_xml_tool_source(GENERAL_MISSING_TOOL_ID_NAME_VERSION)
    run_lint(lint_ctx, general.lint_general, XmlToolSource(tool_source))
    assert 'Tool version is missing or empty.' in lint_ctx.error_messages
    assert 'Tool name is missing or empty.' in lint_ctx.error_messages
    assert 'Tool does not define an id attribute.' in lint_ctx.error_messages
    assert 'Tool specifies an invalid profile version [2109].' in lint_ctx.error_messages


def test_general_whitespace_in_versions_and_names(lint_ctx):
    tool_source = get_xml_tool_source(GENERAL_WHITESPACE_IN_VERSIONS_AND_NAMES)
    run_lint(lint_ctx, general.lint_general, XmlToolSource(tool_source))
    assert "Tool version is pre/suffixed by whitespace, this may cause errors: [ 1.0.1 ]." in lint_ctx.warn_messages
    assert "Tool name is pre/suffixed by whitespace, this may cause errors: [ BWA Mapper ]." in lint_ctx.warn_messages
    assert "Requirement version contains whitespace, this may cause errors: [ 1.2.5 ]." in lint_ctx.warn_messages
    assert "Tool ID contains whitespace - this is discouraged: [bwa tool]." in lint_ctx.warn_messages
    assert "Tool targets 16.01 Galaxy profile." in lint_ctx.valid_messages


def test_general_requirement_without_version(lint_ctx):
    tool_source = get_xml_tool_source(GENERAL_REQUIREMENT_WO_VERSION)
    run_lint(lint_ctx, general.lint_general, XmlToolSource(tool_source))
    assert 'Tool version [1.0.1blah] is not compliant with PEP 440.' in lint_ctx.warn_messages
    assert "Requirement bwa defines no version" in lint_ctx.warn_messages
    assert "Requirement without name found" in lint_ctx.error_messages
    assert "Tool specifies profile version [20.09]." in lint_ctx.valid_messages
    assert "Tool defines an id [bwa_tool]." in lint_ctx.valid_messages
    assert "Tool defines a name [BWA Mapper]." in lint_ctx.valid_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 3
    assert len(lint_ctx.warn_messages) == 2
    assert len(lint_ctx.error_messages) == 1


def test_general_valid(lint_ctx):
    tool_source = get_xml_tool_source(GENERAL_VALID)
    run_lint(lint_ctx, general.lint_general, XmlToolSource(tool_source))
    assert 'Tool defines a version [1.0+galaxy1].' in lint_ctx.valid_messages
    assert "Tool specifies profile version [21.09]." in lint_ctx.valid_messages
    assert "Tool defines an id [valid_id]." in lint_ctx.valid_messages
    assert "Tool defines a name [valid name]." in lint_ctx.valid_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 4
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_help_multiple(lint_ctx):
    tool_source = get_xml_tool_source(HELP_MULTIPLE)
    run_lint(lint_ctx, help.lint_help, tool_source)
    assert 'More than one help section found, behavior undefined.' in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 1


def test_help_absent(lint_ctx):
    tool_source = get_xml_tool_source(HELP_ABSENT)
    run_lint(lint_ctx, help.lint_help, tool_source)
    assert 'No help section found, consider adding a help section to your tool.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_help_empty(lint_ctx):
    tool_source = get_xml_tool_source(HELP_EMPTY)
    run_lint(lint_ctx, help.lint_help, tool_source)
    assert 'Help section appears to be empty.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_help_todo(lint_ctx):
    tool_source = get_xml_tool_source(HELP_TODO)
    run_lint(lint_ctx, help.lint_help, tool_source)
    assert 'Tool contains help section.' in lint_ctx.valid_messages
    assert 'Help contains valid reStructuredText.' in lint_ctx.valid_messages
    assert "Help contains TODO text." in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 2
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_help_invalid_rst(lint_ctx):
    tool_source = get_xml_tool_source(HELP_INVALID_RST)
    run_lint(lint_ctx, help.lint_help, tool_source)
    assert 'Tool contains help section.' in lint_ctx.valid_messages
    assert "Invalid reStructuredText found in help - [<string>:2: (WARNING/2) Inline strong start-string without end-string.\n]." in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 1
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_inputs_no_inputs(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_NO_INPUTS)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found no input parameters.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_inputs_no_inputs_datasource(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_NO_INPUTS_DATASOURCE)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'No input parameters, OK for data sources' in lint_ctx.info_messages
    assert 'display tag usually present in data sources' in lint_ctx.info_messages
    assert 'uihints tag usually present in data sources' in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 3
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_inputs_valid(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_VALID)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert "Found 2 input parameters." in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_inputs_param_name(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_PARAM_NAME)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert "Found 5 input parameters." in lint_ctx.info_messages
    assert 'Param input [2] is not a valid Cheetah placeholder.' in lint_ctx.warn_messages
    assert 'Found param input with no name specified.' in lint_ctx.error_messages
    assert 'Param input with empty name.' in lint_ctx.error_messages
    assert "Param input [param_name] 'name' attribute is redundant if argument implies the same name." in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 2
    assert len(lint_ctx.error_messages) == 2


def test_inputs_param_type(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_PARAM_TYPE)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert "Found 2 input parameters." in lint_ctx.info_messages
    assert 'Param input [valid_name] input with no type specified.' in lint_ctx.error_messages
    assert 'Param input [another_valid_name] with empty type specified.' in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 2


def test_inputs_data_param(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_DATA_PARAM)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert "Found 1 input parameters." in lint_ctx.info_messages
    assert "Param input [valid_name] with no format specified - 'data' format will be assumed." in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_inputs_conditional(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_CONDITIONAL)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 10 input parameters.' in lint_ctx.info_messages
    assert "Conditional without a name" in lint_ctx.error_messages
    assert "Select parameter of a conditional [select] options have to be defined by 'option' children elements." in lint_ctx.error_messages
    assert 'Conditional [cond_wo_param] needs exactly one child <param> found 0' in lint_ctx.error_messages
    assert 'Conditional [cond_w_mult_param] needs exactly one child <param> found 2' in lint_ctx.error_messages
    assert 'Conditional [cond_text] first param should have type="select"' in lint_ctx.error_messages
    assert 'Conditional [cond_boolean] first param of type="boolean" is discouraged, use a select' in lint_ctx.warn_messages
    assert "Conditional [cond_boolean] no truevalue/falsevalue found for when block 'False'" in lint_ctx.warn_messages
    assert 'Conditional [cond_w_optional_select] test parameter cannot be optional="true"' in lint_ctx.warn_messages
    assert 'Conditional [cond_w_multiple_select] test parameter cannot be multiple="true"' in lint_ctx.warn_messages
    assert "Conditional [when_wo_value] when without value" in lint_ctx.error_messages
    assert "Conditional [missing_when] no <when /> block found for select option 'none'" in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 6
    assert len(lint_ctx.error_messages) == 6


def test_inputs_select_incompatible_display(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_SELECT_INCOMPATIBLE_DISPLAY)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 3 input parameters.' in lint_ctx.info_messages
    assert 'Select [radio_select] display="radio" is incompatible with optional="true"' in lint_ctx.error_messages
    assert 'Select [radio_select] display="radio" is incompatible with multiple="true"' in lint_ctx.error_messages
    assert 'Select [checkboxes_select] `display="checkboxes"` is incompatible with `optional="false"`, remove the `display` attribute' in lint_ctx.error_messages
    assert 'Select [checkboxes_select] `display="checkboxes"` is incompatible with `multiple="false"`, remove the `display` attribute' in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 4


def test_inputs_duplicated_options(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_SELECT_DUPLICATED_OPTIONS)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 1 input parameters.' in lint_ctx.info_messages
    assert 'Select parameter [select] has multiple options with the same text content' in lint_ctx.error_messages
    assert 'Select parameter [select] has multiple options with the same value' in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 2


def test_inputs_duplicated_options_with_different_select(lint_ctx):
    tool_source = get_xml_tool_source(SELECT_DUPLICATED_OPTIONS_WITH_DIFF_SELECTED)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_inputs_select_deprections(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_SELECT_DEPRECATIONS)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 3 input parameters.' in lint_ctx.info_messages
    assert "Select parameter [select_do] uses deprecated 'dynamic_options' attribute." in lint_ctx.warn_messages
    assert "Select parameter [select_ff] options uses deprecated 'from_file' attribute." in lint_ctx.warn_messages
    assert "Select parameter [select_fp] options uses deprecated 'from_parameter' attribute." in lint_ctx.warn_messages
    assert "Select parameter [select_ff] options uses deprecated 'transform_lines' attribute." in lint_ctx.warn_messages
    assert "Select parameter [select_fp] options uses deprecated 'options_filter_attribute' attribute." in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 5
    assert not lint_ctx.error_messages


def test_inputs_select_option_definitions(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_SELECT_OPTION_DEFINITIONS)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 6 input parameters.' in lint_ctx.info_messages
    assert "Select parameter [select_noopt] options have to be defined by either 'option' children elements, a 'options' element or the 'dynamic_options' attribute." in lint_ctx.error_messages
    assert "Select parameter [select_noopts] options tag defines no options. Use 'from_dataset', 'from_data_table', or a filter that adds values." in lint_ctx.error_messages
    assert "Select parameter [select_fd_op] options have to be defined by either 'option' children elements, a 'options' element or the 'dynamic_options' attribute." in lint_ctx.error_messages
    assert "Select parameter [select_fd_op] contains multiple options elements." in lint_ctx.error_messages
    assert "Select parameter [select_fd_fdt] options uses 'from_dataset' and 'from_data_table' attribute." in lint_ctx.error_messages
    assert "Select parameter [select_noval_notext] has option without value" in lint_ctx.error_messages
    assert "Select parameter [select_noval_notext] has option without text" in lint_ctx.warn_messages
    assert "Select parameter [select_meta_file_key_incomp] 'meta_file_key' is only compatible with 'from_dataset'." in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert len(lint_ctx.error_messages) == 7


def test_inputs_select_filter(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_SELECT_FILTER)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 1 input parameters.' in lint_ctx.info_messages
    assert "Select parameter [select_filter_types] contains filter without type." in lint_ctx.error_messages
    assert "Select parameter [select_filter_types] contains filter with unknown type 'unknown_filter_type'." in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 2


def test_inputs_validator_incompatibilities(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_VALIDATOR_INCOMPATIBILITIES)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 2 input parameters.' in lint_ctx.info_messages
    assert "Parameter [param_name]: 'in_range' validators are not expected to contain text (found 'TEXT')" in lint_ctx.warn_messages
    assert "Parameter [param_name]: validator with an incompatible type 'in_range'" in lint_ctx.error_messages
    assert "Parameter [param_name]: 'in_range' validators need to define the 'min' or 'max' attribute(s)" in lint_ctx.error_messages
    assert "Parameter [param_name]: attribute 'filename' is incompatible with validator of type 'regex'" in lint_ctx.error_messages
    assert "Parameter [param_name]: expression validators are expected to contain text" in lint_ctx.error_messages
    assert "Parameter [param_name]: '[' is no valid regular expression: unterminated character set at position 0" in lint_ctx.error_messages
    assert "Parameter [another_param_name]: 'metadata' validators need to define the 'check' or 'skip' attribute(s)" in lint_ctx.error_messages
    assert "Parameter [param_name]: 'value_in_data_table' validators need to define the 'table_name' attribute" in lint_ctx.error_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert len(lint_ctx.error_messages) == 7


def test_inputs_validator_correct(lint_ctx):
    tool_source = get_xml_tool_source(INPUTS_VALIDATOR_CORRECT)
    run_lint(lint_ctx, inputs.lint_inputs, tool_source)
    assert 'Found 5 input parameters.' in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_inputs_repeats(lint_ctx):
    tool_source = get_xml_tool_source(REPEATS)
    run_lint(lint_ctx, inputs.lint_repeats, tool_source)
    assert "Repeat does not specify name attribute." in lint_ctx.error_messages
    assert "Repeat does not specify title attribute." in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 2


def test_outputs_missing(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_MISSING)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert 'Tool contains no outputs section, most tools should produce outputs.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_outputs_multiple(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_MULTIPLE)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert '0 outputs found.' in lint_ctx.info_messages
    assert 'Tool contains multiple output sections, behavior undefined.' in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_outputs_unknown_tag(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_UNKNOWN_TAG)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert '0 outputs found.' in lint_ctx.info_messages
    assert 'Unknown element found in outputs [output]' in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_outputs_unnamed_invalid_name(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_UNNAMED_INVALID_NAME)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert '2 outputs found.' in lint_ctx.info_messages
    assert "Tool output doesn't define a name - this is likely a problem." in lint_ctx.warn_messages
    assert "Tool data output with missing name doesn't define an output format." in lint_ctx.warn_messages
    assert 'Tool output name [2output] is not a valid Cheetah placeholder.' in lint_ctx.warn_messages
    assert "Collection output with undefined 'type' found." in lint_ctx.warn_messages
    assert "Tool collection output 2output doesn't define an output format." in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 5
    assert not lint_ctx.error_messages


def test_outputs_format_input(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_FORMAT_INPUT)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert '1 outputs found.' in lint_ctx.info_messages
    assert "Using format='input' on data, format_source attribute is less ambiguous and should be used instead." in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_outputs_collection_format_source(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_COLLECTION_FORMAT_SOURCE)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert "Tool data output 'reverse' should use either format_source or format/ext" in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_outputs_format_action(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_FORMAT_ACTION)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_outputs_discover_tool_provided_metadata(lint_ctx):
    tool_source = get_xml_tool_source(OUTPUTS_DISCOVER_TOOL_PROVIDED_METADATA)
    run_lint(lint_ctx, outputs.lint_output, tool_source)
    assert '1 outputs found.' in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_stdio_default_for_default_profile(lint_ctx):
    tool_source = get_xml_tool_source(STDIO_DEFAULT_FOR_DEFAULT_PROFILE)
    run_lint(lint_ctx, stdio.lint_stdio, XmlToolSource(tool_source))
    assert "No stdio definition found, tool indicates error conditions with output written to stderr." in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_stdio_default_for_nonlegacy_profile(lint_ctx):
    tool_source = get_xml_tool_source(STDIO_DEFAULT_FOR_NONLEGACY_PROFILE)
    run_lint(lint_ctx, stdio.lint_stdio, XmlToolSource(tool_source))
    assert "No stdio definition found, tool indicates error conditions with non-zero exit codes." in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_stdio_multiple_stdio(lint_ctx):
    tool_source = get_xml_tool_source(STDIO_MULTIPLE_STDIO)
    run_lint(lint_ctx, stdio.lint_stdio, XmlToolSource(tool_source))
    assert "More than one stdio tag found, behavior undefined." in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 1


def test_stdio_invalid_child_or_attrib(lint_ctx):
    tool_source = get_xml_tool_source(STDIO_INVALID_CHILD_OR_ATTRIB)
    run_lint(lint_ctx, stdio.lint_stdio, XmlToolSource(tool_source))
    assert "Unknown stdio child tag discovered [reqex]. Valid options are exit_code and regex." in lint_ctx.warn_messages
    assert "Unknown attribute [descriptio] encountered on exit_code tag." in lint_ctx.warn_messages
    assert "Unknown attribute [descriptio] encountered on regex tag." in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 3
    assert not lint_ctx.error_messages


def test_stdio_invalid_match(lint_ctx):
    tool_source = get_xml_tool_source(STDIO_INVALID_MATCH)
    run_lint(lint_ctx, stdio.lint_stdio, XmlToolSource(tool_source))
    assert "Match '[' is no valid regular expression: unterminated character set at position 0" in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 1


def test_tests_absent(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_ABSENT)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert 'No tests found, most tools should define test cases.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


def test_tests_data_source(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_ABSENT_DATA_SOURCE)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert 'No tests found, that should be OK for data_sources.' in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_tests_param_output_names(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_PARAM_OUTPUT_NAMES)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert '1 test(s) found.' in lint_ctx.valid_messages
    assert "Test 1: Found test param tag without a name defined." in lint_ctx.error_messages
    assert "Test 1: Test param non_existent_test_name not found in the inputs" in lint_ctx.error_messages
    assert "Test 1: Found output tag without a name defined." in lint_ctx.error_messages
    assert "Test 1: Found output tag with unknown name [nonexistent_output], valid names [['existent_output']]" in lint_ctx.error_messages
    assert "Test 1: Found output_collection tag without a name defined." in lint_ctx.error_messages
    assert "Test 1: Found output_collection tag with unknown name [nonexistent_collection], valid names [['existent_collection']]" in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 1
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 6


def test_tests_expect_failure_output(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_EXPECT_FAILURE_OUTPUT)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert 'No valid test(s) found.' in lint_ctx.warn_messages
    assert "Test 1: Cannot specify outputs in a test expecting failure." in lint_ctx.error_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert len(lint_ctx.error_messages) == 1


def test_tests_without_expectations(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_WO_EXPECTATIONS)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert 'Test 1: No outputs or expectations defined for tests, this test is likely invalid.' in lint_ctx.warn_messages
    assert 'No valid test(s) found.' in lint_ctx.warn_messages
    assert not lint_ctx.info_messages
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 2
    assert not lint_ctx.error_messages


def test_tests_valid(lint_ctx):
    tool_source = get_xml_tool_source(TESTS_VALID)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert "1 test(s) found." in lint_ctx.valid_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 1
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_tests_asserts(lint_ctx):
    tool_source = get_xml_tool_source(ASSERTS)
    run_lint(lint_ctx, tests.lint_tsts, tool_source)
    assert "Test 1: unknown assertion 'invalid'" in lint_ctx.error_messages
    assert "Test 1: unknown attribute 'invalid_attrib' for 'has_text'" in lint_ctx.error_messages
    assert "Test 1: missing attribute 'text' for 'has_text'" in lint_ctx.error_messages
    assert "Test 1: attribute 'value' for 'has_size' needs to be 'int' got '500k'" in lint_ctx.error_messages
    assert "Test 1: attribute 'delta' for 'has_size' needs to be 'int' got '1O'" in lint_ctx.error_messages
    assert "Test 1: unknown attribute 'invalid_attrib_also_checked_in_nested_asserts' for 'not_has_text'" in lint_ctx.error_messages
    assert "Test 1: 'has_size' needs to specify 'n', 'min', or 'max'" in lint_ctx.error_messages
    assert "Test 1: 'has_n_columns' needs to specify 'n', 'min', or 'max'" in lint_ctx.error_messages
    assert "Test 1: 'has_n_lines' needs to specify 'n', 'min', or 'max'" in lint_ctx.error_messages
    assert not lint_ctx.warn_messages
    assert len(lint_ctx.error_messages) == 9


def test_xml_order(lint_ctx):
    tool_source = get_xml_tool_source(XML_ORDER)
    run_lint(lint_ctx, xml_order.lint_xml_order, tool_source)
    assert 'Unknown tag [wrong_tag] encountered, this may result in a warning in the future.' in lint_ctx.info_messages
    assert 'Best practice violation [stdio] elements should come before [command]' in lint_ctx.warn_messages
    assert len(lint_ctx.info_messages) == 1
    assert not lint_ctx.valid_messages
    assert len(lint_ctx.warn_messages) == 1
    assert not lint_ctx.error_messages


COMPLETE = """<tool>
    <macros>
        <import>macros.xml</import>
        <xml name="test_macro">
            <param name="select" type="select">
                <option value="a">a</option>
                <option value="a">a</option>
            </param>
        </xml>
    </macros>
    <inputs>
        <expand macro="test_macro"/>
        <param/>
        <expand macro="test_macro2"/>
    </inputs>
</tool>
"""

COMPLETE_MACROS = """<macros>
    <xml name="test_macro2">
        <param name="No_type"/>
    </xml>
</macros>
"""


def test_tool_and_macro_xml(lint_ctx_xpath, lint_ctx):
    """
    test linters (all of them via lint_tool_source_with) on a tool and macro xml file
    checking a list of asserts, where each assert is a 4-tuple:
    - expected message
    - filename suffix (tool.xml or macro.xml)
    - line number (1 based)
    - xpath
    """
    with tempfile.TemporaryDirectory() as tmp:
        tool_path = os.path.join(tmp, "tool.xml")
        macros_path = os.path.join(tmp, "macros.xml")
        with open(tool_path, "w") as tmpf:
            tmpf.write(COMPLETE)
        with open(macros_path, "w") as tmpf:
            tmpf.write(COMPLETE_MACROS)
        tool_xml, _ = load_with_references(tool_path)

    tool_source = XmlToolSource(tool_xml)
    # lint once with the lint context using XMLLintMessageXPath and XMLLintMessageLine
    lint_tool_source_with(lint_ctx_xpath, tool_source)
    lint_tool_source_with(lint_ctx, tool_source)

    asserts = (
        ("Select parameter [select] has multiple options with the same value", 5, "/tool/inputs/param[1]"),
        ("Found param input with no name specified.", 13, "/tool/inputs/param[2]"),
        ("Param input [No_type] input with no type specified.", 3, "/tool/inputs/param[3]")
    )
    for a in asserts:
        message, line, xpath = a
        found = False
        for lint_message in lint_ctx_xpath.message_list:
            if lint_message.message != message:
                continue
            found = True
            assert lint_message.xpath == xpath, f"Assumed xpath {xpath}; found xpath {lint_message.xpath} for: {message}"
        assert found, f"Did not find {message}"
        for lint_message in lint_ctx.message_list:
            if lint_message.message != message:
                continue
            found = True
            assert lint_message.line == line, f"Assumed line {line}; found line {lint_message.line} for: {message}"
        assert found, f"Did not find {message}"


# TODO COPIED from test/unit/app/tools/test_tool_deserialization.py
CWL_TOOL = """
cwlVersion: v1.0
class: CommandLineTool
baseCommand: echo
inputs:
  message:
    type: string
    inputBinding:
      position: 1
outputs: []
"""

YAML_TOOL = """
class: GalaxyTool
id: simple_constructs_y
name: simple_constructs_y
version: 1.0
command:
  >
    echo "$booltest"  >> $out_file1;
inputs:
- name: booltest
  type: boolean
  truevalue: booltrue
  falsevalue: boolfalse
  checked: false
outputs:
  out_file1:
    format: txt
"""


def test_linting_yml_tool(lint_ctx):
    with tempfile.TemporaryDirectory() as tmp:
        tool_path = os.path.join(tmp, "tool.yml")
        with open(tool_path, "w") as tmpf:
            tmpf.write(YAML_TOOL)
        tool_sources = load_tool_sources_from_path(tmp)
        assert len(tool_sources) == 1, "Expected 1 tool source"
        tool_source = tool_sources[0][1]
        lint_tool_source_with(lint_ctx, tool_source)
    assert "Tool defines a version [1.0]." in lint_ctx.valid_messages
    assert "Tool defines a name [simple_constructs_y]." in lint_ctx.valid_messages
    assert "Tool defines an id [simple_constructs_y]." in lint_ctx.valid_messages
    assert "Tool specifies profile version [16.04]." in lint_ctx.valid_messages
    assert not lint_ctx.info_messages
    assert len(lint_ctx.valid_messages) == 4
    assert not lint_ctx.warn_messages
    assert not lint_ctx.error_messages


def test_linting_cwl_tool(lint_ctx):
    with tempfile.TemporaryDirectory() as tmp:
        tool_path = os.path.join(tmp, "tool.cwl")
        with open(tool_path, "w") as tmpf:
            tmpf.write(CWL_TOOL)
        tool_sources = load_tool_sources_from_path(tmp)
        assert len(tool_sources) == 1, "Expected 1 tool source"
        tool_source = tool_sources[0][1]
        lint_tool_source_with(lint_ctx, tool_source)
    assert "Tool defines a version [0.0.1]." in lint_ctx.valid_messages
    assert "Tool defines a name [tool]." in lint_ctx.valid_messages
    assert "Tool defines an id [tool]." in lint_ctx.valid_messages
    assert "Tool specifies profile version [16.04]." in lint_ctx.valid_messages
    assert "CWL appears to be valid." in lint_ctx.info_messages
    assert "Description of tool is empty or absent." in lint_ctx.warn_messages
    assert "Tool does not specify a DockerPull source." in lint_ctx.warn_messages
    assert "Modern CWL version [v1.0]." in lint_ctx.info_messages
    assert len(lint_ctx.info_messages) == 2
    assert len(lint_ctx.valid_messages) == 4
    assert len(lint_ctx.warn_messages) == 2
    assert not lint_ctx.error_messages



import pytest

from galaxy.tool_util.deps.mulled.mulled_build import (
    base_image_for_targets,
    build_target,
    DEFAULT_BASE_IMAGE,
    DEFAULT_EXTENDED_BASE_IMAGE,
    mull_targets,
    target_str_to_targets
)
from ..util import external_dependency_management


@pytest.mark.parametrize("target,version,base_image", [
    ('maker', None, DEFAULT_EXTENDED_BASE_IMAGE),
    ('qiime', '1.9.1', DEFAULT_EXTENDED_BASE_IMAGE),
    ('samtools', None, DEFAULT_BASE_IMAGE),
])
@external_dependency_management
def test_base_image_for_targets(target, version, base_image):
    target = build_target(target, version=version)
    assert base_image_for_targets([target]) == base_image


@external_dependency_management
def test_mulled_build_files_cli(tmpdir):
    singularity_image_dir = tmpdir.mkdir('singularity image dir')
    target = build_target('zlib')
    mull_targets([target], command='build-and-test', singularity=True, singularity_image_dir=singularity_image_dir)
    assert singularity_image_dir.join('zlib').exists()


def test_target_str_to_targets():
    target_str = 'samtools=1.3.1--4,bedtools=2.22'
    targets = target_str_to_targets(target_str)
    assert (targets[0].package, targets[0].version, targets[0].build) == ('samtools', '1.3.1', '4')
    assert (targets[1].package, targets[1].version, targets[1].build) == ('bedtools', '2.22', None)



from subprocess import CalledProcessError

import pytest

from galaxy.tool_util.deps.container_classes import DOCKER_CONTAINER_TYPE
from galaxy.tool_util.deps.container_resolvers.mulled import (
    CachedMulledDockerContainerResolver,
    CachedMulledSingularityContainerResolver,
    MulledDockerContainerResolver,
)
from galaxy.tool_util.deps.containers import ContainerRegistry
from galaxy.tool_util.deps.dependencies import (
    AppInfo,
    ToolInfo,
)
from galaxy.tool_util.deps.requirements import ToolRequirement


SINGULARITY_IMAGES = ('foo:1.0--bar', 'baz:2.22', 'mulled-v2-fe8a3b846bc50d24e5df78fa0b562c43477fe9ce:9f946d13f673ab2903cb0da849ad42916d619d18-0')


@pytest.fixture
def container_registry():
    app_info = AppInfo(
        involucro_auto_init=True,
        enable_mulled_containers=True,
        container_image_cache_path=".",
    )
    return ContainerRegistry(app_info)


def test_container_registry(container_registry, mocker):
    mocker.patch('galaxy.tool_util.deps.mulled.util._get_namespace', return_value=["samtools"])
    tool_info = ToolInfo(requirements=[ToolRequirement(name="samtools", version="1.10", type="package")])
    container_description = container_registry.find_best_container_description(
        [DOCKER_CONTAINER_TYPE],
        tool_info,
        install=False,
    )
    assert container_description.type == "docker"
    assert "samtools:1.10" in container_description.identifier


def test_docker_container_resolver_detects_docker_cli_absent(mocker):
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.which', return_value=None)
    resolver = CachedMulledDockerContainerResolver()
    assert resolver._cli_available is False


def test_docker_container_resolver_detects_docker_cli(mocker):
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled', return_value='/bin/docker')
    resolver = CachedMulledDockerContainerResolver()
    assert resolver.cli_available


def test_cached_docker_container_docker_cli_absent_resolve(mocker):
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.which', return_value=None)
    resolver = CachedMulledDockerContainerResolver()
    assert resolver.cli_available is False
    assert resolver.resolve(enabled_container_types=[], tool_info={}) is None


def test_docker_container_docker_cli_absent_resolve(mocker):
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.which', return_value=None)
    resolver = MulledDockerContainerResolver()
    assert resolver.cli_available is False
    requirement = ToolRequirement(name="samtools", version="1.10", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.targets_to_mulled_name', return_value='samtools:1.10--h2e538c0_3')
    container_description = resolver.resolve(enabled_container_types=['docker'], tool_info=tool_info)
    assert container_description.type == 'docker'
    assert container_description.identifier == 'quay.io/biocontainers/samtools:1.10--h2e538c0_3'


def test_docker_container_docker_cli_exception_resolve(mocker):
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.which', return_value='/bin/docker')
    resolver = MulledDockerContainerResolver()
    assert resolver.cli_available is True
    requirement = ToolRequirement(name="samtools", version="1.10", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.targets_to_mulled_name', return_value='samtools:1.10--h2e538c0_3')
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.docker_cached_container_description', side_effect=CalledProcessError(1, 'bla'))
    container_description = resolver.resolve(enabled_container_types=['docker'], tool_info=tool_info, install=True)
    assert resolver.cli_available is True
    assert container_description.type == 'docker'
    assert container_description.identifier == 'quay.io/biocontainers/samtools:1.10--h2e538c0_3'


def test_cached_singularity_container_resolver_uncached(mocker):
    mocker.patch('os.listdir', return_value=SINGULARITY_IMAGES)
    mocker.patch('os.path.exists', return_value=True)
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.safe_makedirs')
    resolver = CachedMulledSingularityContainerResolver(app_info=mocker.Mock(container_image_cache_path='/'))
    requirement = ToolRequirement(name="foo", version="1.0", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    container_description = resolver.resolve(enabled_container_types=['singularity'], tool_info=tool_info)
    assert container_description.type == 'singularity'
    assert container_description.identifier == '/singularity/mulled/foo:1.0--bar'


def test_cached_singularity_container_resolver_dir_mtime_cached(mocker):
    mocker.patch('os.listdir', return_value=SINGULARITY_IMAGES)
    mocker.patch('os.path.exists', return_value=True)
    mocker.patch('galaxy.tool_util.deps.container_resolvers.mulled.safe_makedirs')
    mocker.patch('os.stat', return_value=mocker.Mock(st_mtime=42))
    resolver = CachedMulledSingularityContainerResolver(app_info=mocker.Mock(container_image_cache_path='/'), cache_directory_cacher_type='dir_mtime')
    requirement = ToolRequirement(name="baz", version="2.22", type="package")
    tool_info = ToolInfo(requirements=[requirement])
    container_description = resolver.resolve(enabled_container_types=['singularity'], tool_info=tool_info)
    assert container_description.type == 'singularity'
    assert container_description.identifier == '/singularity/mulled/baz:2.22'
    requirement = ToolRequirement(name="foo", version="1.0", type="package")
    tool_info.requirements.append(requirement)
    container_description = resolver.resolve(enabled_container_types=['singularity'], tool_info=tool_info)
    assert container_description.type == 'singularity'
    assert container_description.identifier == '/singularity/mulled/mulled-v2-fe8a3b846bc50d24e5df78fa0b562c43477fe9ce:9f946d13f673ab2903cb0da849ad42916d619d18-0'



"""This module contains a linting functions for tool inputs."""
import re

from galaxy.util import string_as_bool
from ._util import (
    is_datasource,
    is_valid_cheetah_placeholder
)
from ..parser.util import _parse_name

FILTER_TYPES = [
    'data_meta',
    'param_value',
    'static_value',
    'regexp',
    'unique_value',
    'multiple_splitter',
    'attribute_value_splitter',
    'add_value',
    'remove_value',
    'sort_by',
]

ATTRIB_VALIDATOR_COMPATIBILITY = {
    "check": ["metadata"],
    "expression": ["substitute_value_in_message"],
    "table_name": ["dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table", "value_in_data_table", "value_not_in_data_table"],
    "filename": ["dataset_metadata_in_file"],
    "metadata_name": ["dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table", "dataset_metadata_in_file"],
    "metadata_column": ["dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table", "value_in_data_table", "value_not_in_data_table", "dataset_metadata_in_file"],
    "line_startswith": ["dataset_metadata_in_file"],
    "min": ["in_range", "length", "dataset_metadata_in_range"],
    "max": ["in_range", "length", "dataset_metadata_in_range"],
    "exclude_min": ["in_range", "dataset_metadata_in_range"],
    "exclude_max": ["in_range", "dataset_metadata_in_range"],
    "split": ["dataset_metadata_in_file"],
    "skip": ["metadata"]
}

PARAMETER_VALIDATOR_TYPE_COMPATIBILITY = {
    "integer": ["in_range", "expression"],
    "float": ["in_range", "expression"],
    "data": ["metadata", "unspecified_build", "dataset_ok_validator", "dataset_metadata_in_range", "dataset_metadata_in_file", "dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table", "expression"],
    "data_collection": ["metadata", "unspecified_build", "dataset_ok_validator", "dataset_metadata_in_range", "dataset_metadata_in_file", "dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table", "expression"],
    "text": ["regex", "length", "empty_field", "value_in_data_table", "value_not_in_data_table", "expression"],
    "select": ["no_options", "regex", "length", "empty_field", "value_in_data_table", "value_not_in_data_table", "expression"],
    "drill_down": ["no_options", "regex", "length", "empty_field", "value_in_data_table", "value_not_in_data_table", "expression"],
    "data_column": ["no_options", "regex", "length", "empty_field", "value_in_data_table", "value_not_in_data_table", "expression"]
}


def lint_inputs(tool_xml, lint_ctx):
    """Lint parameters in a tool's inputs block."""
    datasource = is_datasource(tool_xml)
    inputs = tool_xml.findall("./inputs//param")
    # determine node to report for general problems with outputs
    tool_node = tool_xml.find("./inputs")
    if tool_node is None:
        tool_node = tool_xml.getroot()
    num_inputs = 0
    for param in inputs:
        num_inputs += 1
        param_attrib = param.attrib
        if "name" not in param_attrib and "argument" not in param_attrib:
            lint_ctx.error("Found param input with no name specified.", node=param)
            continue
        param_name = _parse_name(param_attrib.get("name"), param_attrib.get("argument"))
        if "name" in param_attrib and "argument" in param_attrib:
            if param_attrib.get("name") == _parse_name(None, param_attrib.get("argument")):
                lint_ctx.warn(f"Param input [{param_name}] 'name' attribute is redundant if argument implies the same name.", node=param)
        if param_name.strip() == "":
            lint_ctx.error("Param input with empty name.", node=param)
        elif not is_valid_cheetah_placeholder(param_name):
            lint_ctx.warn(f"Param input [{param_name}] is not a valid Cheetah placeholder.", node=param)

        # TODO lint for params with duplicated name (in inputs & outputs)

        if "type" not in param_attrib:
            lint_ctx.error(f"Param input [{param_name}] input with no type specified.", node=param)
            continue
        elif param_attrib["type"].strip() == "":
            lint_ctx.error(f"Param input [{param_name}] with empty type specified.", node=param)
            continue
        param_type = param_attrib["type"]

        # TODO lint for valid param type - attribute combinations

        if param_type == "data":
            if "format" not in param_attrib:
                lint_ctx.warn(f"Param input [{param_name}] with no format specified - 'data' format will be assumed.", node=param)
        elif param_type == "select":
            # get dynamic/statically defined options
            dynamic_options = param.get("dynamic_options", None)
            options = param.findall("./options")
            filters = param.findall("./options/filter")
            select_options = param.findall('./option')

            if dynamic_options is not None:
                lint_ctx.warn(f"Select parameter [{param_name}] uses deprecated 'dynamic_options' attribute.", node=param)

            # check if options are defined by exactly one possibility
            if param.getparent().tag != "conditional":
                if (dynamic_options is not None) + (len(options) > 0) + (len(select_options) > 0) != 1:
                    lint_ctx.error(f"Select parameter [{param_name}] options have to be defined by either 'option' children elements, a 'options' element or the 'dynamic_options' attribute.", node=param)
            else:
                if len(select_options) == 0:
                    lint_ctx.error(f"Select parameter of a conditional [{param_name}] options have to be defined by 'option' children elements.", node=param)

            # lint dynamic options
            if len(options) == 1:
                filters = options[0].findall("./filter")
                # lint filters
                # TODO check if dataset is available for filters referring other datasets
                filter_adds_options = False
                for f in filters:
                    ftype = f.get("type", None)
                    if ftype is None:
                        lint_ctx.error(f"Select parameter [{param_name}] contains filter without type.", node=f)
                        continue
                    if ftype not in FILTER_TYPES:
                        lint_ctx.error(f"Select parameter [{param_name}] contains filter with unknown type '{ftype}'.", node=f)
                        continue
                    if ftype in ['add_value', 'data_meta']:
                        filter_adds_options = True
                    # TODO more linting of filters

                from_file = options[0].get("from_file", None)
                from_parameter = options[0].get("from_parameter", None)
                from_dataset = options[0].get("from_dataset", None)
                from_data_table = options[0].get("from_data_table", None)
                # TODO check if input param is present for from_dataset

                if (from_file is None and from_parameter is None
                        and from_dataset is None and from_data_table is None
                        and not filter_adds_options):
                    lint_ctx.error(f"Select parameter [{param_name}] options tag defines no options. Use 'from_dataset', 'from_data_table', or a filter that adds values.", node=options[0])

                for deprecated_attr in ["from_file", "from_parameter", "options_filter_attribute", "transform_lines"]:
                    if options[0].get(deprecated_attr) is not None:
                        lint_ctx.warn(f"Select parameter [{param_name}] options uses deprecated '{deprecated_attr}' attribute.", node=options[0])

                if from_dataset is not None and from_data_table is not None:
                    lint_ctx.error(f"Select parameter [{param_name}] options uses 'from_dataset' and 'from_data_table' attribute.", node=options[0])

                if options[0].get("meta_file_key", None) is not None and from_dataset is None:
                    lint_ctx.error(f"Select parameter [{param_name}] 'meta_file_key' is only compatible with 'from_dataset'.", node=options[0])

            elif len(options) > 1:
                lint_ctx.error(f"Select parameter [{param_name}] contains multiple options elements.", node=options[1])

            # lint statically defined options
            if any('value' not in option.attrib for option in select_options):
                lint_ctx.error(f"Select parameter [{param_name}] has option without value", node=param)
            if any(option.text is None for option in select_options):
                lint_ctx.warn(f"Select parameter [{param_name}] has option without text", node=param)

            select_options_texts = list()
            select_options_values = list()
            for option in select_options:
                value = option.attrib.get("value", "")
                if option.text is None:
                    text = value.capitalize()
                else:
                    text = option.text
                select_options_texts.append((text, option.attrib.get("selected", "false")))
                select_options_values.append((value, option.attrib.get("selected", "false")))
            if len(set(select_options_texts)) != len(select_options_texts):
                lint_ctx.error(f"Select parameter [{param_name}] has multiple options with the same text content", node=param)
            if len(set(select_options_values)) != len(select_options_values):
                lint_ctx.error(f"Select parameter [{param_name}] has multiple options with the same value", node=param)

            multiple = string_as_bool(param_attrib.get("multiple", "false"))
            optional = string_as_bool(param_attrib.get("optional", multiple))
            if param_attrib.get("display") == "checkboxes":
                if not multiple:
                    lint_ctx.error(f'Select [{param_name}] `display="checkboxes"` is incompatible with `multiple="false"`, remove the `display` attribute', node=param)
                if not optional:
                    lint_ctx.error(f'Select [{param_name}] `display="checkboxes"` is incompatible with `optional="false"`, remove the `display` attribute', node=param)
            if param_attrib.get("display") == "radio":
                if multiple:
                    lint_ctx.error(f'Select [{param_name}] display="radio" is incompatible with multiple="true"', node=param)
                if optional:
                    lint_ctx.error(f'Select [{param_name}] display="radio" is incompatible with optional="true"', node=param)
        # TODO: Validate type, much more...

        # lint validators
        # TODO check if dataset is available for validators referring other datasets
        validators = param.findall("./validator")
        for validator in validators:
            vtype = validator.attrib['type']
            if param_type in PARAMETER_VALIDATOR_TYPE_COMPATIBILITY:
                if vtype not in PARAMETER_VALIDATOR_TYPE_COMPATIBILITY[param_type]:
                    lint_ctx.error(f"Parameter [{param_name}]: validator with an incompatible type '{vtype}'", node=validator)
            for attrib in ATTRIB_VALIDATOR_COMPATIBILITY:
                if attrib in validator.attrib and vtype not in ATTRIB_VALIDATOR_COMPATIBILITY[attrib]:
                    lint_ctx.error(f"Parameter [{param_name}]: attribute '{attrib}' is incompatible with validator of type '{vtype}'", node=validator)
            if vtype == "expression":
                if validator.text is None:
                    lint_ctx.error(f"Parameter [{param_name}]: expression validators are expected to contain text", node=validator)
                else:
                    try:
                        re.compile(validator.text)
                    except Exception as e:
                        lint_ctx.error(f"Parameter [{param_name}]: '{validator.text}' is no valid regular expression: {str(e)}", node=validator)
            if vtype not in ["expression", "regex"] and validator.text is not None:
                lint_ctx.warn(f"Parameter [{param_name}]: '{vtype}' validators are not expected to contain text (found '{validator.text}')", node=validator)
            if vtype in ["in_range", "length", "dataset_metadata_in_range"] and ("min" not in validator.attrib and "max" not in validator.attrib):
                lint_ctx.error(f"Parameter [{param_name}]: '{vtype}' validators need to define the 'min' or 'max' attribute(s)", node=validator)
            if vtype in ["metadata"] and ("check" not in validator.attrib and "skip" not in validator.attrib):
                lint_ctx.error(f"Parameter [{param_name}]: '{vtype}' validators need to define the 'check' or 'skip' attribute(s)", node=validator)
            if vtype in ["value_in_data_table", "value_not_in_data_table", "dataset_metadata_in_data_table", "dataset_metadata_not_in_data_table"] and "table_name" not in validator.attrib:
                lint_ctx.error(f"Parameter [{param_name}]: '{vtype}' validators need to define the 'table_name' attribute", node=validator)

    conditional_selects = tool_xml.findall("./inputs//conditional")
    for conditional in conditional_selects:
        conditional_name = conditional.get('name')
        if not conditional_name:
            lint_ctx.error("Conditional without a name", node=conditional)
        if conditional.get("value_from"):
            # Probably only the upload tool use this, no children elements
            continue
        first_param = conditional.findall("param")
        if len(first_param) != 1:
            lint_ctx.error(f"Conditional [{conditional_name}] needs exactly one child <param> found {len(first_param)}", node=conditional)
            continue
        first_param = first_param[0]
        first_param_type = first_param.get('type')
        if first_param_type == 'boolean':
            lint_ctx.warn(f'Conditional [{conditional_name}] first param of type="boolean" is discouraged, use a select', node=first_param)
        elif first_param_type != 'select':
            lint_ctx.error(f'Conditional [{conditional_name}] first param should have type="select"', node=first_param)
            continue

        if first_param_type == 'select':
            select_options = _find_with_attribute(first_param, 'option', 'value')
            option_ids = [option.get('value') for option in select_options]
        else:  # boolean
            option_ids = [
                first_param.get('truevalue', 'true'),
                first_param.get('falsevalue', 'false')
            ]

        for incomp in ["optional", "multiple"]:
            if string_as_bool(first_param.get(incomp, False)):
                lint_ctx.warn(f'Conditional [{conditional_name}] test parameter cannot be {incomp}="true"', node=first_param)

        whens = conditional.findall('./when')
        if any('value' not in when.attrib for when in whens):
            lint_ctx.error(f"Conditional [{conditional_name}] when without value", node=conditional)

        when_ids = [w.get('value') for w in whens if w.get('value') is not None]

        for option_id in option_ids:
            if option_id not in when_ids:
                lint_ctx.warn(f"Conditional [{conditional_name}] no <when /> block found for {first_param_type} option '{option_id}'", node=conditional)

        for when_id in when_ids:
            if when_id not in option_ids:
                if first_param_type == 'select':
                    lint_ctx.warn(f"Conditional [{conditional_name}] no <option /> found for when block '{when_id}'", node=conditional)
                else:
                    lint_ctx.warn(f"Conditional [{conditional_name}] no truevalue/falsevalue found for when block '{when_id}'", node=conditional)

    if datasource:
        # TODO only display is subtag of inputs, uihints is a separate top level tag (supporting only attrib minwidth)
        for datasource_tag in ('display', 'uihints'):
            if not any(param.tag == datasource_tag for param in inputs):
                lint_ctx.info(f"{datasource_tag} tag usually present in data sources", node=tool_node)

    if num_inputs:
        lint_ctx.info(f"Found {num_inputs} input parameters.", node=tool_node)
    else:
        if datasource:
            lint_ctx.info("No input parameters, OK for data sources", node=tool_node)
        else:
            lint_ctx.warn("Found no input parameters.", node=tool_node)


def lint_repeats(tool_xml, lint_ctx):
    """Lint repeat blocks in tool inputs."""
    repeats = tool_xml.findall("./inputs//repeat")
    for repeat in repeats:
        if "name" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify name attribute.", node=repeat)
        if "title" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify title attribute.", node=repeat)


def _find_with_attribute(element, tag, attribute, test_value=None):
    rval = []
    for el in (element.findall(f'./{tag}') or []):
        if attribute not in el.attrib:
            continue
        value = el.attrib[attribute]
        if test_value is not None:
            if value == test_value:
                rval.append(el)
        else:
            rval.append(el)
    return rval



"""Logic for dealing with cwltool as an optional dependency.

Use this as the import interface for cwltool and just call
:func:`ensure_cwltool_available` before using any of the imported
functionality at runtime.
"""
import re
import warnings

warnings.filterwarnings("ignore", message=r"[\n.]DEPRECATION: Python 2", module="cwltool")

import requests

try:
    from cwltool import (
        main,
        workflow,
        job,
        process,
        pathmapper,
    )
except ImportError:
    main = None  # type: ignore[assignment]
    workflow = None  # type: ignore[assignment]
    job = None  # type: ignore[assignment]
    process = None  # type: ignore[assignment]
    pathmapper = None  # type: ignore[assignment]

try:
    from cwltool.context import (
        getdefault,
        LoadingContext,
        RuntimeContext,
    )
    from cwltool.job import relink_initialworkdir
    from cwltool.stdfsaccess import StdFsAccess
except ImportError:
    getdefault = None  # type: ignore[assignment]
    LoadingContext = None  # type: ignore[assignment,misc]
    relink_initialworkdir = None  # type: ignore[assignment]
    RuntimeContext = None  # type: ignore[assignment,misc]
    StdFsAccess = None  # type: ignore[assignment,misc]

try:
    from cwltool import load_tool
    from cwltool.load_tool import (
        default_loader,
        resolve_and_validate_document,
    )
except ImportError:
    default_loader = None  # type: ignore[assignment]
    load_tool = None  # type: ignore[assignment]
    resolve_and_validate_document = None  # type: ignore[assignment]


try:
    from cwltool import command_line_tool
except ImportError:
    command_line_tool = None  # type: ignore[assignment]

try:
    from cwltool.load_tool import resolve_and_validate_document
except ImportError:
    resolve_and_validate_document = None  # type: ignore[assignment]

try:
    import shellescape
except ImportError:
    shellescape = None

try:
    import schema_salad
    from schema_salad import (
        ref_resolver,
        sourceline,
    )
except ImportError:
    schema_salad = None  # type: ignore[assignment]
    ref_resolver = None  # type: ignore[assignment]
    sourceline = None  # type: ignore[assignment]

needs_shell_quoting = re.compile(r"""(^$|[\s|&;()<>\'"$@])""").search

# if set to True, file format checking is not performed.
beta_relaxed_fmt_check = True


def ensure_cwltool_available():
    """Assert optional dependencies proxied via this module are available at runtime.

    Throw an ImportError with a description of the problem if they do not exist.
    """
    if main is None or workflow is None or shellescape is None:
        message = "This feature requires cwltool and dependencies to be available, they are not."
        if main is None:
            message += " cwltool is not unavailable."
        elif resolve_and_validate_document is None:
            message += " cwltool.load_tool.resolve_and_validate_document is unavailable - cwltool version is too old."
        if requests is None:
            message += " Library 'requests' unavailable."
        if shellescape is None:
            message += " Library 'shellescape' unavailable."
        if schema_salad is None:
            message += " Library 'schema_salad' unavailable."
        raise ImportError(message)


__all__ = (
    'default_loader',
    'ensure_cwltool_available',
    'getdefault',
    'load_tool',
    'LoadingContext',
    'main',
    'needs_shell_quoting',
    'pathmapper',
    'process',
    'ref_resolver',
    'relink_initialworkdir',
    'resolve_and_validate_document',
    'RuntimeContext',
    'schema_salad',
    'shellescape',
    'sourceline',
    'StdFsAccess',
    'workflow',
)
